%% Compare the Random Alpha PageRank model with PageRank
% On a web graph, the random alpha PageRank model yields slightly different
% results from the PageRank model itself.  In this experiment, we analyze
% these differences in terms of the Kendall-tau ordering metric between the
% vectors induced by each method.

%% Setup the experiment
% This experiment should be run from the rapr/experiments/comparison directory
cwd = pwd;
dirtail = 'experiments/comparison'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('rapr:dir','%s should be executed from rapr/%s\n',mfilename,dirtail);
end
addpath('../../matlab'); % ensure we have the RAPr codes available
addpath('../../libbvg'); % ensure we have the libbvg codes available

%%
% Type a copy of ourselves so we have a record
type driver_compare_with_pagerank.m

%%
% Setup the standard list of random variable parameters from each method
params = [
      0,   0, 0.6, 0.9   % uniform [0,6,0.9]
      2,  16,   0,   1   % skewed right
      1,   1, 0.1, 0.9   % equicentric
   -0.5,-0.5, 0.2, 0.7   % bi-modal 
];
nds=size(params,1);

%%
% Setup a list of graphs to test
datadir = '../../data';
matdata = {'uk-2006-05.hostgraph','uk-2007-05.hostgraph'};
bvgdata = {'eu-2005','nz2006','uk2005','us2004scc','indochina-2004','enwiki-20080103'};   
% Use smaller versions to test...
%matdata = {'uk-2006-05.hostgraph'};
%bvgdata = {'eu-2005'};   

% Test that we can load each graph
for i=1:length(matdata), load([datadir filesep matdata{i} '.mat']); end
for i=1:length(bvgdata), A=bvgraph([datadir filesep bvgdata{i}]); end

cellfun(@(n) fprintf('%s ...\n',n), [matdata bvgdata]);
fprintf('... files successfully tested!\n');

%% Evaluate PageRank using GQ
% To evaluate the RAPr model functions, we use the GQ algorithm.  

%%
% Use an N point quadrature rule.  Our tests showed this gives about
% 10^-7 accuracy on all the distributions in the test
N=33; tol=1e-9;

%%
% Reset the results state
results=[];

%% 
% Run the tests for all the Matlab graphs
for i=1:length(matdata)
    name=matdata{i}; load([datadir filesep name '.mat']);
    for di=1:nds
        a=params(di,1);b=params(di,2); l=params(di,3); r=params(di,4);
        d=alphadist('beta',a,b,l,r); eA=d.mf(1); eA=eA(end);
        ri.graph=name; ri.d = d; ri.eA = eA; % ri is the current result
        tic; [ri.ex ri.stdx ri.xw]=gqrapr(P,N,d,tol); ri.dtgq=toc;
        tic; [ri.xeA flag ri.prconv] = inoutpr(P,eA,[],tol); ri.dtpr=toc;
        results=[results ri]; save 'compare-results.mat' results % save results
        fprintf('%20s : beta(%4g,%4g,%4g,%4g) gq(%2i)=%7f sec, pow(EA)=%7f sec\n',...
            name,a,b,l,r,N,ri.dtgq,ri.dtpr);
    end
end

%%
% Run the tests for all the bvggraphs
for i=1:length(bvgdata)
    name=bvgdata{i}; A=bvgraph([datadir filesep bvgdata{i}]); A=sparse(A);
    P = normout(A);
    for di=1:nds
        a=params(di,1);b=params(di,2); l=params(di,3); r=params(di,4);
        d=alphadist('beta',a,b,l,r); eA=d.mf(1); eA=eA(end);
        ri.graph=name; ri.d = d; ri.eA = eA; % ri is the current result
        tic; [ri.ex ri.stdx ri.xw]=gqrapr(P,N,d,tol); ri.dtgq=toc;
        tic; [ri.xeA flag ri.prconv] = inoutpr(P,eA,[],tol); ri.dtpr=toc;
        results=[results ri]; save 'compare-results.mat' results % save results
        fprintf('%20s : beta(%4g,%4g,%4g,%4g) gq(%2i)=%7f sec, pow(EA)=%7f sec\n',...
            name,a,b,l,r,N,ri.dtgq,ri.dtpr);
    end
end
