%% Evaluate the convergence of our algorithms for a larger problem
% We have four codes that should be computing identical quantities in the
% limit as some computational resource parameter goes to infinity.  In this
% experiment, we test to make sure all the algorithms converge as we
% increase the computational resourses supplied.  Convergence is measured
% as the change in the solution when adding a bit more computational power.
%
% This code expands on driver_convergence_large.m by plotting time instead
% of iteration counts and reducing the pce tolerance to 1e-5;

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end

%%
% Add the necessary path
addpath('../../libbvg');
addpath('../../matlab');

%%
% Load and convert the graph
A = bvgraph('../../data/cnr-2000');
A = sparse(A); P = normout(A);

%% Show convergence of each algorithm
% We'll use a series of distributions and show how each algorithm
% converges.  Then, for each algorithm and each distribution, we will look
% at how much each solution changes from one step to the next.  Again, we
% use the same distributions we've been using throughout the paper.

Nmc=1e3; Npd=2e4; gqNs=0:10:40; pceNs=gqNs;
mcalg=1e-8;gqalg=1e-8;pcealg='ujacobi';pcetol=1e-5;
du1 = alphadist('unif',0.6,0.9);           % uniform [0.6,0.9]
db1 = alphadist('beta',2,16,0,1);          % skewed right
db2 = alphadist('beta',1,1,0.1,0.9);       % equicentric
db3 = alphadist('beta',-0.5,-0.5,0.2,0.7); % bi-modal
ds = {du1,db1,db2,db3}; nds = length(ds);
results = []; [results(1:nds).d] = deal(ds{:}); % intialize results

%%
% Compute a monte-carlo approximation and store the convergence results
tic;
for di=1:nds
    [ex stdx alphas conv]=mcrapr(P,Nmc,results(di).d,mcalg);
    results(di).ex.mc=ex; results(di).stdx.mc=stdx; results(di).conv.mc=conv;
end
fprintf('... %7i monte carlo samples took %f seconds\n', Nmc, toc);
save 'converge-large-results-time.mat' results P % save intermediate results

%%
% Compute a path damping approximation and store the convergence results
tic;
for di=1:nds
    [ex stdx pdcoeffs conv] = pdrapr(P,results(di).d,0,Npd);
    results(di).ex.pd=ex; results(di).stdx.pd=stdx; results(di).conv.pd=conv;
end
fprintf('... %7i path damping iterations took %f seconds\n', Npd, toc);
save 'converge-large-results-time.mat' results P % save intermediate results

%%
% Compute a seris of GQ approximations and explicitly compute the changes
% that result.
tic
for di=1:nds
    conv=zeros(length(gqNs),4); d=results(di).d;
    for Ni=1:length(gqNs)
        tic; [ex stdx] = gqrapr(P,gqNs(Ni)+1,d,gqalg); dt=toc;
        [ex1 stdx1] = gqrapr(P,gqNs(Ni)+2,d,gqalg); 
        conv(Ni,:) = [gqNs(Ni),norm(ex-ex1,1),norm(stdx-stdx1,1),dt];
    end
    results(di).ex.gq=ex; results(di).stdx.gq=stdx; results(di).conv.gq=conv;
end
fprintf('... %i GQPR solves took %f seconds\n',sum(gqNs)+sum(gqNs+1),toc);
save 'converge-large-results-time.mat' results P % save intermediate results

%%
% Compute a seris of PCE approximations and explicitly compute the changes
% that result.  (Note that this will be identical to the GQ results.)
tic
for di=1:nds
    conv=zeros(length(pceNs),4); d=results(di).d;
    for Ni=1:length(pceNs)
        tic; [ex stdx] = pcerapr(P,pceNs(Ni),d,pcealg,pcetol); dt=toc; 
        [ex1 stdx1] = pcerapr(P,pceNs(Ni)+1,d,pcealg,pcetol); 
        conv(Ni,:) = [pceNs(Ni),norm(ex-ex1,1),norm(stdx-stdx1,1),dt];
    end
    results(di).ex.pce=ex; results(di).stdx.pce=stdx; results(di).conv.pce=conv;
end
fprintf('... %i PCEPR solves took %f seconds\n',sum(pceNs)+sum(pceNs+1),toc);

%%
% Save the results
save 'converge-large-results-time.mat' results P

%% Plot the convergence results by algorithm.
% Each distribution is drawn with its native colors from previous plots
% with points indicated by '.' for an expectation and '+' for a standard
% deviation.  We do not es
algs={'mc','pd','gq','pce'}; nas=length(algs);  
ls={'LineWidth',1.2};ms={'MarkerSize',15}; ao={'defaultaxeslinewidth',1.0};
s={'-.','-','--',':'}; c={'k','k','k','k'};
% c={[1 0.5 0.5],[0.5 0.5 1],[0 0.75 0.5],[1 0 0.25]};
for ai=1:nas
    figure(ai); set(gcf,'Color','w','defaultaxesfontsize',16,ao{:}); clf;
    for di=1:nds
        cd=results(di).conv.(algs{ai}); x=cd(:,end); y1=cd(:,2);
        pfun='loglog'; 
        if size(cd,2)>3, h=feval(pfun,x,y1,['',s{di}],x,cd(:,3),['',s{di}]);
        else h=feval(pfun,x,y1,['',s{di}]); end, set(h(1),ms{:});
        hold on; set(h,'Color',c{di}); pbaspect([2.5 1 1]); ylim([1e-16 1]);
        box off; print(gcf,sprintf('converge-large-time-%s.eps',algs{ai}),'-depsc2');
    end
end

%%
% Plot just the beta(2,16) (the hard case) from each algorithm to show
% scaling behavior.
algs={'mc','pd','gq','pce'}; nas=length(algs);  
ls={'LineWidth',1.0};ms={'MarkerSize',15}; ao={'defaultaxeslinewidth',0.8};
s={'-.','-','--',':'}; c={':','--','-','-.'};
di=2; figure(1); set(gcf,'Color','w','defaultaxesfontsize',12,ao{:}); clf;
hold on;
for ai=1:nas
    cd=results(di).conv.(algs{ai}); x=cd(:,end); y1=cd(:,2);
    loglog(x,y1,['k' c{ai}],ls{:}); box on;
end
set(gca,'XScale','log'); set(gca,'YScale','log'); pbaspect([2.75 1 1]);
legend('Monte Carlo','Path Damping','Quadrature','Polynomial Chaos',...
    'Location','SW'); legend boxoff; xlabel('Time (sec)');
print(gcf,sprintf('converge-large-time.eps'),'-depsc2','-cmyk');    