%% Evaluate the correctness of our algorithm
% We have four codes that should be computing identical quantities in the
% limit as some computational resource parameter goes to infinity.  In this
% experiment, we test all the codes against each other and show that we get
% the expected behavior.  This experiment does not show we are always
% computing the right answer, but it shows that four independent
% implementations of similar quantities converge to similar answers
% (in this case, the differences begin to approximate the order of the
% machine precision.)

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
%%
% Some code revisions for initial testing

% just use the largest strong component for this experiment;
%P = Pcc;
%load ('../../data/example-small.mat');

%%
% Add the necessary path
addpath('../../matlab');

%% Show convergence of each algorithm
% We'll use a series of distributions and show how each algorithm
% converges.  Then, for each algorithm and each distribution, we will look
% at how much each solution changes from one step to the next.
Nmc=1e5; Npd=Nmc;
db1 = alphadist('beta',2,16,0,1);
db2 = alphadist('beta',3,2,0.1,0.9);
du1 = alphadist('unif',0,1);
du2 = alphadist('unif',0.2,0.75);
ds = {db1,db2,du1,du2}; nds = length(ds);
[results(1:nds).d] = deal(ds{:});

%%
% Compute a monte-carlo approximation and store the convergence results
for di=1:nds
    [ex stdx alphas conv]=mcrapr(P,Nmc,results(di).d,'direct');
    results(di).ex.mc=ex; results(di).stdx.mc=stdx; results(di).conv.mc=conv;
end

%%
% Compute a path damping approximation and store the convergence results
for di=1:nds
    [ex stdx pdcoeffs conv] = pdrapr(P,results(di).d,0,Npd);
    results(di).ex.pd=ex; results(di).stdx.pd=stdx; results(di).conv.pd=conv;
end

%%
% Compute a seris of GQ approximations and explicitly compute the changes
% that result.
gqNs=0:10:100;
for di=1:nds
    conv=zeros(length(gqNs),3);
    for Ni=1:length(gqNs)
        [ex stdx] = gqrapr(P,gqNs(Ni)+1,results(di).d,'direct');
        [ex1 stdx1] = gqrapr(P,gqNs(Ni)+2,results(di).d,'direct');
        conv(Ni,:) = [gqNs(Ni),norm(ex-ex1,1),norm(stdx-stdx1,1)];
    end
    results(di).ex.gq=ex; results(di).stdx.gq=stdx; results(di).conv.gq=conv;
end

%%
% Compute a seris of PCE approximations and explicitly compute the changes
% that result.  (Note that this will be identical to the GQ results.)
pceNs=gqNs;
for di=1:nds
    conv=zeros(length(pceNs),3);
    for Ni=1:length(pceNs)
        [ex stdx] = pcerapr(P,pceNs(Ni),results(di).d,'direct');
        [ex1 stdx1] = pcerapr(P,pceNs(Ni)+1,results(di).d,'direct');
        conv(Ni,:) = [pceNs(Ni),norm(ex-ex1,1),norm(stdx-stdx1,1)];
    end
    results(di).ex.pce=ex; results(di).stdx.pce=stdx; results(di).conv.pce=conv;
end

%%
% Save the results
save 'converge-results.mat' results P

%% Plot the convergence results
% 
algs={'mc','pd','gq','pce'}; nas=length(algs);
for di=1:nds
    for ai=1:nas
        subplot(nds,nas,(ai-1)*nds+di); 
        cd=results(di).conv.(algs{ai}); x=cd(:,1); y1=cd(:,2);
        if log10(max(x))>2.5, pfun='loglog'; else pfun='semilogy'; end
        if size(cd,2)>2, 
            feval(pfun,x,y1,'.-',x,cd(:,3),'*-');
        else
            feval(pfun,x,y1,'.-');
        end
    end
end
