%% Accurate RAPr computations
% In this experiment, we will compute the expectation of our RAPr model to
% machine precision taking advantage of known properties of Gauss
% Quadrature (GQ).  First, an N-point GQ rule is accurate for all
% polynomials of degree (2N).  Given a matrix P of size n x n, the PageRank
% vector is a  a polynomial function of degree n in alpha, the random
% parameter.  Consequently, using an n/2 point GQ rule will integrate our
% function exactly.  In this experiment, we construct such a rule and
% compute the expectation.

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
% just use the largest strong component for this experiment;
P = Pcc;
%load ('../../data/example-small.mat');
n=size(P,1);

%%
% For this experiment, we'll assume a Beta(2,16) distribution
a=2; b=16;

%% Compute the GQ rule
% Our first step is to compute an n/2 point GQ rule.  We need to be careful
% at this step to make sure we get the tridiagonal matrix of Jacobi
% coefficients correct for the three-term recurrence.  As a first step,
% we'll try Gautschi's code.

N = ceil(n/2);
ab = r_jacobi01(N,a,b);
d = alphadist('beta',a,b);

%%
% To check the coefficients', we will plot them.  
figure(1);
h=semilogx(1:N,ab(:,1),'--',1:N,ab(:,2),'-'); box off; xlim([0, 2e3]); 
pbaspect([3,1,1]); set(h(1),'Color',[1 0.5 0.5]);set(h(2),'Color',[0.5 0.5 1]);
legend('\alpha_k', '\beta_k'); text(2e3,1/16,'1/16'); text(2e3,1/2,'1/2');

%%
% Now we need to get the points and weights.
xw=gauss(N,ab); xw(:,2) = xw(:,2)/beta(b+1,a+1); 

%%
% Make a high resolution plot
figure(2);
stem(xw(:,1),xw(:,2),'LineWidth',0.01,'MarkerSize',0.2,'Color','k'); box off;
pbaspect([5 1 1]); print(gcf,'-dpdf','-r1200','-PaperSizeB0','gq-rule.pdf');

%%
% Check the sum of the weights.
sum(xw(:,2))

%%
% Unfortunately, the sum is not zero, which means we are not computing the
% function entirely accurately.  However, this is about the order of
% machine precision, so we should be okay.

%% Solve all the PageRank systems
% To solve all the systems efficiently, we're going to use the Hessenberg
% form of the matrix P.  This allows us to update the factor alpha without
% recomputing a factorization of the matrix.

[Q,H] = hess(full(P'));
v = ones(n,1)./n; w = Q'*v;

X=zeros(n,N);
for i=1:N
    tic; X(:,i) = solve_pr_hess(n,Q,H,xw(i,1),w); dt=toc;
    fprintf('alpha = %6.5f; min(x) = %8e, sum(x) = %8e, norm(x,1) = %8e; dt=%6f\n',...
        xw(i,1),min(X(:,i)),sum(X(:,i)),norm(X(:,i),1), dt);
end

save 'exact-gq-solves.mat' X xw P

%%
% Form the GQ solution
eex = X*xw(:,2);

%% Compare against Path Damping
% Thankfully, path damping is very easy to implement.  We are going to
% check how well our solutions compare against the exact solution.
Npd = 2e5;
m = d.mf(Npd);
Ptkx = v; ex=zeros(n,1); t=0; hpd=zeros(Npd,3); s=0;
for k=1:Npd
    ex = ex + (m(k)-m(k+1)).*Ptkx; t=1-norm(ex,1); s=s+(m(k)-m(k+1));
    y = P'*(Ptkx); Ptkx = y + (1-sum(y))*v;
    z1 = ex + t*Ptkx; z2 = ex + (1-s)*Ptkx;
    hpd(k,1) = norm(eex-ex,1); hpd(k,2)=norm(eex-z1,1); hpd(k,3)=norm(eex-z2,1);
    if mod(k,1000)==0, fprintf('k = %6i ; erry = %8e ; errz = %8e\n',k,t,1-s); end
end

%%
% Save the results
save 'exact-pd-hist.mat' eex m hpd P

%% 
% Plot the results
loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
    1:Npd,hpd(:,3),'-.',1:Npd,1-m(1)+m(2:end),'.:');
