%% Accurate RAPr computations with chebfuns

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
% just use the largest strong component for this experiment;
P = Pcc; A = Acc;
%load ('../../data/example-small.mat');
n=size(P,1);

%%
% Add the path 
addpath('../../matlab');

%% 
% Factorize the matrix once.  We are going to compute the solutions for
% many values of alpha.  The computationally efficient means of updating
% alpha is to use an upper hessenberg decomposition of the matrix P.  This
% is real and always exists.  Then the solution can be computed in O(n^2)
% time for any value a, with perfect numerical stability.
[Q,H] = hess(full(P'));
v = ones(n,1)./n; w = Q'*v;

%%
% Add the chebfun system.  Trefethen's chebfun system allows us to
% represent functions via their Chebyshev polynomial expansion on the
% interval from [-1 1].  We will use this system to compute our integrals,
% which should give a machine accurate representation.
addpath('~/dev/matlab/download/chebfun');

%% Generate the weight function
% With the chebfun system, can generate the weight function for the beta
% distribution by generating a variable x and modifying it with the
% equation of the weight function, or we can specify it directly as an
% inline function.  We've taken the first choice to show off a bit of the
% chebfun system's flexiblity in Matlab.

a=2;
b=16;
r=0.9999;
d=alphadist('beta',a,b,0,r);

x = chebfun('x');
f = (((x+1)./2).^a.*(1-(x+1)./2).^b)/(2*beta(a+1,b+1));

%%
% Now let's check to make sure we got our weight correct.  The sum function
% on a chebfun produces the definite integral.

fprintf('int( weight(x), 0, 1 ) = %g\n', sum(f));

%% Construct the PageRank polynomial
% The PageRank vector is a polynomial of degree n in alpha.  After
% scaling it by the weight function, (a polynomial of degree (a+b)), we
% find % the overall degree of our polynomial is n+a+b+1 (or maybe a+b+n).

N = 2*(n+a+b+1);

%%
% Unfortunately, building the PageRank polynomial for all components is
% somewhat inefficient with the chebfun system.  It doesn't support vectors
% of functions quite as well as I would like.  To make things slightly
% efficient, we will write a memoized PageRank solution method.

type memo_solve_pr_hess.m

%%
% This function returns another function that memoizes a PageRank solution
% vector so it doesn't have to every evaluate the same value of alpha
% twice.  The solve_pr_hess function doesn't do anything special.

type solve_pr_hess.m

%%
% With these functions, we can construct the PageRank polynomials. 
xa = memo_solve_pr_hess(Q,H,w,n);
x = cell(n,1);
ex = zeros(n,1);
for i=1:n
    xif = @(x) xa(i,r.*(x+1)./2).*((r.*(x+1)./2).^b.*(r-r.*(x+1)./2).^a.*(r./2))/exp((a+b+1)*log(r)+betaln(a+1,b+1));
    xi = chebfun(xif,N);
    x{i} = xi;
    ex(i) = sum(xi);
end

save 'exact-chebfuns.mat' xa x ex P

eex = ex;


%% Compare against Path Damping
% Thankfully, path damping is very easy to implement.  We are going to
% check how well our solutions compare against the exact solution.
Npd = 2e3;
m = d.mf(Npd);
Ptkx = v; ex=zeros(n,1); t=0; hpd=zeros(Npd,5); s=0; ex1=ex; ex2=ex;
for k=1:Npd
    ex = ex + (m(k)-m(k+1)).*Ptkx; t=norm(ex,1); s=s+(m(k)-m(k+1)); r=m(1)-m(k+1);
    ex1 = ex1 + m(k).*Ptkx; ex2 = ex2 + m(k+1).*Ptkx;
    y = P'*(Ptkx); Ptkx = y + (1-sum(y))*v;
    z1 = ex + (1-t)*Ptkx; z2 = ex + (1-s)*Ptkx; z3=ex./t; z4=ex1-ex2 + m(k+1).*Ptkx;
    hpd(k,1) = norm(eex-ex,1); hpd(k,2)=norm(eex-z1,1); 
    hpd(k,3)=norm(eex-z2,1); hpd(k,4)=norm(eex-z3,1); hpd(k,5)=norm(eex-z4,1);
    if mod(k,1000)==0, fprintf('k = %6i ; erry = %8e ; errz = %8e\n',k,t,1-s); end
end

save 'exact-pd-hist.mat' eex m hpd P

%% 
% Plot the results
% loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
%     1:Npd,hpd(:,3),'-.',1:Npd,1-m(1)+m(2:end),'.:');
loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
    1:Npd,hpd(:,3),'-.',1:Npd,hpd(:,4),'.:', ...
    1:Npd,hpd(:,5),'--');

%% Test against GQ
%

gqNs=0:10:100;
gqconv=zeros(length(gqNs),2);
for Ni=1:length(gqNs)
    N=gqNs(Ni);
    [ex stdx] = gqrapr(P,N+1,d,'direct');
    %[ex1 stdx1] = gqrapr(P,N+2,db,'direct');
    gqconv(Ni,:) = [N,norm(ex-eex,1)];
end
semilogy(gqconv(:,1),gqconv(:,2),'.-');
