%% Accurate RAPr computations with Fejer quadrature
% To get an accurate RAPr vector, we are going to try using a Fejer rule
% with a many points.  A Fejer rule is like a Clenshaw-Curtis rule, but
% excludes the endpoints, and hence, permits singularities.  

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
% just use the largest strong component for this experiment;
P = Pcc; A = Acc;
%load ('../../data/example-small.mat');
n=size(P,1);

%%
% Add the path 
addpath('../../matlab');

%% Generate the weight function
% With the chebfun system, can generate the weight function for the beta
% distribution by generating a variable x and modifying it with the
% equation of the weight function, or we can specify it directly as an
% inline function.  We've taken the first choice to show off a bit of the
% chebfun system's flexiblity in Matlab.

a=2;
b=16;
r=1;
d=alphadist('beta',a,b,0,r);

%% Construct the Fejer rule.
% The PageRank vector is a rational of degree n in alpha (I think).  After
% scaling it by the weight function, (a polynomial of degree (a+b)), we
% find the overall degree of our rational function is n+a+b.  We'll build a
% rule to integrate polynomials of up to degree 

N = 2*(n+a+b+1);

%% 
[fx fw] = fejer2_rule_compute(N);

%%
X = zeros(n,length(fx));
Pt = P'; tic; olddt=0;
v = ones(n,1)./n;
bw = zeros(N,1);
for i=1:N
    ti = fx(i);
    alpha = 0.5*(fx(i)+1);
    bw(i)=((r.*(ti+1)./2).^b.*(r-r.*(ti+1)./2).^a.*(r./2))/exp((a+b+1)*log(r)+betaln(a+1,b+1));
    X(:,i)=(speye(n)-alpha*Pt)\((1-alpha)*v);
    dt=toc; etr=N*dt/i-dt; 
    if dt>olddt+1,
        fprintf('... solve %5i a=%7.5f time=%7.2f sec left=%7.2f sec \n',...
            i,alpha,toc,etr);    
        olddt = dt;
    end
end

save 'exact-fejer-2.mat' fx fw P X

eex = X*(bw.*fw);


%% Compare against Path Damping
% Thankfully, path damping is very easy to implement.  We are going to
% check how well our solutions compare against the exact solution.
Npd = 2e5;
m = d.mf(Npd);
Ptkx = v; ex=zeros(n,1); t=0; hpd=zeros(Npd,5); s=0; ex1=ex; ex2=ex; tic;
for k=1:Npd
    ex = ex + (m(k)-m(k+1)).*Ptkx; t=norm(ex,1); s=s+(m(k)-m(k+1)); r=m(1)-m(k+1);
    ex1 = ex1 + m(k).*Ptkx; ex2 = ex2 + m(k+1).*Ptkx;
    y = P'*(Ptkx); Ptkx = y + (1-sum(y))*v;
    z1 = ex + (1-t)*Ptkx; z2 = ex + (1-s)*Ptkx; z3=ex./t; z4=ex1-ex2 + m(k+1).*Ptkx;
    hpd(k,1) = norm(eex-ex,1); hpd(k,2)=norm(eex-z1,1); 
    hpd(k,3)=norm(eex-z2,1); hpd(k,4)=norm(eex-z3,1); hpd(k,5)=norm(eex-z4,1);
    dt=toc; etr=Npd*dt/k-dt;
    if mod(k,1000)==0, fprintf('k = %6i ; erry = %8e ; errz = %8e ; etr=%7f sec\n',...
            k,t,1-s,etr); end
end

save 'exact-pd-hist.mat' eex m hpd P

%% 
% Plot the results
% loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
%     1:Npd,hpd(:,3),'-.',1:Npd,1-m(1)+m(2:end),'.:');
loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
    1:Npd,hpd(:,3),'-.',1:Npd,hpd(:,4),'.:', ...
    1:Npd,hpd(:,5),'--');

%% Test against GQ
%

gqNs=0:10:100;
gqconv=zeros(length(gqNs),2);
for Ni=1:length(gqNs)
    N=gqNs(Ni);
    [ex stdx] = gqrapr(P,N+1,d,'direct');
    %[ex1 stdx1] = gqrapr(P,N+2,db,'direct');
    gqconv(Ni,:) = [N,norm(ex-eex,1)];
end
semilogy(gqconv(:,1),gqconv(:,2),'.-');
