%% Accurate RAPr computations with Path Damping
% To get an accurate RAPr vector, we are going to try using a Fejer rule
% with a many points.  A Fejer rule is like a Clenshaw-Curtis rule, but
% excludes the endpoints, and hence, permits singularities.  

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
% just use the largest strong component for this experiment;
P = Pcc; A = Acc;
%load ('../../data/example-small.mat');
n=size(P,1);
v=ones(n,1)./n;

%%
% Add the path 
addpath('../../matlab');

%% Generate the distribution
a=2;
b=16;
r=1;
d=alphadist('beta',a,b,0,r);

%% Solve the system with path damping
% With path damping, we are going to compute the solution until the norm no
% longer increases.  We are going to use compensated summation due to the
% large number of terms in this sequence.

Npd = 2e6;
m = d.mf(Npd);
Ptkx = v; t=0; tic; oldt=0; s=0; e=0; y=0; temp=0;
for k=1:Npd
    % compute ex as s with compensated summation
    temp = s; y=(m(k)-m(k+1)).*Ptkx+e; s=temp+y; e=(temp-s)+y;
    t=norm(s,1); delta=t-oldt; oldt=t;
    y = P'*(Ptkx);
    if m(k)-m(k+1)<1e-20, s=s+e; rnex=s+Ptkx*(1-norm(s,1)); break; end
    dt=toc; etr=Npd*dt/k-dt;
    if mod(k,100000)==0, fprintf('k = %7i ; deltat = %10e ; pd = %10e, etr=%7f sec\n',...
            k,delta,m(k)-m(k+1),etr); end
end

eex = rnex;
save 'exact-pd-ex.mat' eex m P ex

%% 
% Plot the results
% loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
%     1:Npd,hpd(:,3),'-.',1:Npd,1-m(1)+m(2:end),'.:');
loglog(1:Npd,hpd(:,1),'-',1:Npd,hpd(:,2),'--',...
    1:Npd,hpd(:,3),'-.',1:Npd,hpd(:,4),'.:', ...
    1:Npd,hpd(:,5),'--');

%% Test against GQ
%

gqNs=0:10:100;
gqconv=zeros(length(gqNs),2);
for Ni=1:length(gqNs)
    N=gqNs(Ni);
    [ex stdx] = gqrapr(P,N+1,d,'direct');
    %[ex1 stdx1] = gqrapr(P,N+2,db,'direct');
    gqconv(Ni,:) = [N,norm(ex-eex,1)];
end
semilogy(gqconv(:,1),gqconv(:,2),'.-');
