%% Accurate RAPr computations with the Symbolic Toolbox
% Our initial attempt at using the symbolic toolbox may have been slightly
% incorrect.  This code attempts to use the symbolic 

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/harvard500.mat'); 
% just use the largest strong component for this experiment;
P = Pcc; A = Acc;
%load ('../../data/example-small.mat'); A(1,:) = ones(1,size(A,1));
n=size(P,1);
v=ones(n,1)./n;

%%
% Setup _where_ to do the integrals
symexact = 0;

%%
% Add the path 
addpath('../../matlab');

%% Generate the linear system
% The goal with this experiment is to symbolically solve the PageRank
% system in the value of alpha for a non-trivial case.  To do this, we have
% to be a little careful with how we write everything.  (For the harvard500
% matrix, this isn't explicitly necessary, but this code was originally
% written for the case when the matrix was a little bit bigger.

%%
% Generate our symbolic variable alpha
clear alpha;
alpha=sym('alpha','real');

%%
% Generate the right hand side and the base matrix for the PageRank vector.
e = sym(ones(n,1));
b = (1-alpha)*e/n;
fprintf('... build identity matrix ... '); tic
M = diag(e);
fprintf('... done! %f sec\n', toc);

%%
% Setup the PageRank matrix as a symbolic matrix.
fprintf('... assembling Pagerank matrix ...\n'); tic
[nzi nzj nzv] = find(A);
d = sum(A,2); nextupdate = floor(nnz(A)/100);
for i=1:nnz(A)
    M(nzj(i),nzi(i))=M(nzj(i),nzi(i))-alpha/d(nzi(i));
    if i > nextupdate, 
        fprintf('... ... %i/%i in %7f sec\n', i, nnz(A),toc); 
        nextupdate=nextupdate+floor(nnz(A)/100);
    end
end

%%
% Solve the symbolic system and save the output
fprintf('... solving symbolic system ...\n'); tic
x = M\b;
fprintf('... done! %f sec\n', toc);

save 'exact-sym-inv.mat' x

%% Write the functions for Mathematica
% For the har500 system, we cannot integrate the expressions explicitly, so
% we export them to Mathematica for highly accurate (32 digit) numerical
% integration.  I really wish that Matlab's symbolic toolbox had a way to
% do this.
syms a;
xa = subs(x,alpha,a);
fid=fopen('exact-sym-inv.mf','wt');
for i=1:n, fprintf(fid,'%s\n',char(xa(i))); end
fclose(fid);
clear a;


%% Integrate expressions exactly
% For some smaller systems, this function can integrate the results itself.
% So if symexact is set, then it'll solve systems here.  Otherwise, it'll
% load the results from Mathematica.  (Which means you have to go in and
% compute the results with Mathematica separately.)

% Setup our standard list of parameters
params = [
      0,   0, 0.6, 0.9   % uniform [0,6,0.9]
      2,  16,   0,   1   % skewed right
      1,   1, 0.1, 0.9   % equicentric
   -0.5,-0.5, 0.2, 0.7   % bi-modal 
];

if symexact
for i=1:size(params,1)
    a=params(i,1);b=params(i,2); l=params(i,3); r=params(i,4);
    results(i).d = alphadist('beta',a,b,l,r);
    f = (alpha-sym(l))^b*(sym(r)-alpha)^a;
    %wf = int(f,alpha,l,r);
    wf = (r-l)^(a+b+1)*beta(a+1,b+1);
    % computed un-normalized ex and ex2
    ex = zeros(n,1); ex2=ex; dpts=unique(round(linspace(1,n,100))); dp=1; tic;
    for j=1:n
        ex(j) = double(int(x(j)*f,alpha,l,r));
        ex2(j) = double(int(x(j)*f,alpha,l,r));
        if j>=dpts(dp)
            dt=toc; etr=n*dt/j; 
            fprintf('dist=%2i node=%6i etr=%7f sec\n',i,j,etr);
        end
    end
    ex = ex./double(wf);
    ex2 = ex2./double(wf);
    stdx = sqrt(ex2-ex.^2);
    exactex(:,i) = ex;
    exactstdx(:,i) = stdx;
    results(i).ex.exact = ex;
    results(i).stdx.exact = stdx;
end
end

%% Load results from files
if ~symexact,
for i=1:size(params,1)
    a=params(i,1);b=params(i,2); l=params(i,3); r=params(i,4);
    results(i).d = alphadist('beta',a,b,l,r);
    mathematicadata = struct2cell(load(sprintf('har500-b%i.ex.mat',i)));
    results(i).ex.exact = mathematicadata{1};
    mathematicadata = struct2cell(load(sprintf('har500-b%i.ex2.mat',i)));
    results(i).ex2.exact = mathematicadata{1};
    results(i).stdx.exact = sqrt(results(i).ex2.exact-results(i).ex.exact.^2);
end
end
   
%% Compare against algorithms
% Now, let's see how all the algorithms to.  We'll use the same parameters
% from the convergence computations.
Nmc=1e5; Npd=Nmc; gqNs=0:10:100; pceNs=gqNs;
mcalg='direct';pcealg='direct';gqalg='direct';

%%
% Compute a monte-carlo approximation and store the convergence results.
% The mcrapr code takes an additional (final) parameter which allows us to
% check intermediate convergence against an exact solution.  If this is
% set, then convergence results are reported with reference to these exact
% solutions instead of the local change in solution.  
tic;
for di=1:length(results)
    eex = results(di).ex.exact; estdx = results(di).stdx.exact; d = results(di).d;
    [ex stdx alphas conv]=mcrapr(P,Nmc,d,mcalg,[],[eex estdx]);
    results(di).conv.mc=conv;
end
fprintf('... %7i monte carlo samples took %f seconds\n', Nmc, toc);
save 'exact-har500-conv-results.mat' results % save intermediate results

%%
% Compute a path damping approximation and store the convergence results.
% Path damping also takes the extra exact solutions.  
tic;
for di=1:length(results)
    eex = results(di).ex.exact; estdx = results(di).stdx.exact; d = results(di).d;
    [ex stdx pdcoeffs conv] = pdrapr(P,results(di).d,0,Npd,[],[eex estdx]);
    results(di).conv.pd=conv;
end
fprintf('... %7i path damping iterations took %f seconds\n', Npd, toc);
save 'exact-har500-conv-results.mat' results % save intermediate results

%%
% Compute a series of GQ approximations and explicitly difference in the
% result.
tic;
for di=1:length(results)
    eex = results(di).ex.exact; estdx = results(di).stdx.exact; d = results(di).d;
    conv=zeros(length(gqNs),3);
    for Ni=1:length(gqNs)
        [ex stdx] = gqrapr(P,gqNs(Ni)+1,d,gqalg);
        conv(Ni,:) = [gqNs(Ni),norm(ex-eex,1),norm(stdx-estdx,1)];
    end
    results(di).conv.gq=conv;
    %figure(di); semilogy(conv(:,1),conv(:,2),'.-',conv(:,1),conv(:,3),'*--');
end
fprintf('... %i GQPR solves took %f seconds\n',sum(gqNs)+sum(gqNs+1),toc);
save 'exact-har500-conv-results.mat' results % save intermediate results

%%
% Compute a seris of PCE approximations and explicitly compute the changes
% in the result.  (Note that this should be identical to the GQ results, but
% ends of changing a little due to numerical accuracy issues near 10^{-16}.)
tic
for di=1:length(results)
    eex = results(di).ex.exact; estdx = results(di).stdx.exact; d = results(di).d;
    conv=zeros(length(pceNs),3); 
    for Ni=1:length(pceNs)
        [ex stdx] = pcerapr(P,pceNs(Ni),d,pcealg);
        conv(Ni,:) = [pceNs(Ni),norm(ex-eex,1),norm(stdx-estdx,1)];
    end
    results(di).conv.pce=conv;
end
fprintf('... %i PCEPR solves took %f seconds\n',sum(pceNs)+sum(pceNs+1),toc)

%%
% Save the results
save 'exact-har500-conv-results.mat' results % save intermediate results

%% Plot the convergence results by algorithm.
% Each distribution is drawn with its native colors from previous plots
% with points indicated by '.' for an expectation and '+' for a standard
% deviation.  We do not es
algs={'mc','pd','gq','pce'}; nas=length(algs); ls={'LineWidth',0.8};ms={'MarkerSize',15};
s={'-.','-','--',':'}; c={[1 0.5 0.5],[0.5 0.5 1],[0 0.75 0.5],[1 0 0.25]};
for ai=1:nas
    figure(ai); set(gcf,'Color','w','defaultaxesfontsize',12); clf;
    for di=1:length(results)
        cd=results(di).conv.(algs{ai}); x=cd(:,1); y1=cd(:,2);
        if log10(max(x))>2.5, pfun='loglog'; else pfun='semilogy'; end
        if size(cd,2)>2, h=feval(pfun,x,y1,['.',s{di}],x,cd(:,3),['+',s{di}]);
        else h=feval(pfun,x,y1,['.',s{di}]); end, set(h(1),ms{:});
        hold on; set(h,'Color',c{di},ls{:}); pbaspect([2.5 1 1]); ylim([1e-16 1]);
        box off; print(gcf,sprintf('exact-har500-conv-%s.eps',algs{ai}),'-depsc2');
    end
end
