function driver_csum_test
%% Test compensated summation algorithm vs. Matlab's summation algorithm.
% While testing our C++ implementations of PageRank, we found problems with
% the norm of the output PageRank vector.  Namely, the vector did not sum
% to one at sufficient precision to make us happy.  Following the advice of
% Wills and Ipsen, we implemented a re-normalized Power/Jacobi algorithm.
% 
% The accuracy of the new algorithm was dramatically worse.  Roundoff errors
% when computing the sum of even 10,000 numbers caused the problem.  Taking
% heed of a subsequent suggestion in Wills and Ipsen led us to implement
% compensated summation, see Kahan summation algorith [Wikipedia] or 
% Chapter 4 in Accuracy and Stability of Numerical Algorithms [Higham].
% The compensated summation algorithm uses a clever floating point ``trick''
% to maintain an estimate of the round-off error and then compensate the sum
% for the accumulated roundoff at each step.  Using this algorithm requires 
% storing one extra floating point number to compute the sum, a rather modest
% cost.  
% 
% Our own codes now gave us full floating point accuracy, i.e. our output
% had a sum of 1.0000000000000000.  When we loaded the values into Matlab
% and computed the norm or sum, we got another value!
%
% This caused us concern about the accuracy of the binary storage and load
% scheme we use for computing derived quanties.  This experiment tests the
% sum of vectors loaded into Matlab using the same compensated summation 
% algorithm as used in the C++ code.

%% Experimental setup
% This experiment should be run from the rapr/experiments/csum directory
cwd = pwd;
dirtail = 'experiments/csum'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end

%% Compute a PageRank vector with bvpr
% All the sums for the inner-outer algorithm and the power algorithm use
% the compensated summation, so let's just use the power algorithm.

!../../libbvg/bvpr ../../data/wb-cs.stanford cs-stanford power

% now load the binary output date
fid=fopen('cs-stanford.pr'); pr=fread(fid,inf,'double'); fclose(fid);

%% Check basics of the PageRank vector
% The PageRank vector should be strictly positive and all elements should
% be less than one.

fprintf('       min(pr) = %18.16e\n', min(pr));
fprintf('       max(pr) = %18.16e\n', max(pr));
fprintf('\n');

%% Check the sum of the PageRank vector
% As we stated, the Matlab sum function gives us slightly different
% answers.  The following test compares sum and csum of the PageRank
% vector.

fprintf('       sum(pr) = %18.16e\n', sum(pr));
fprintf('      norm(pr) = %18.16e\n', norm(pr,1));
fprintf('\n');

%%
% So the Matlab norm and sum do not agree with the output of our code.
% Incidentally, nor do they agree with each other on a strictly positive
% vector.
%
% Let's check our csum algorithm, which is identical to the one implemented
% inside the C++ bvpr code.

type csum.m

fprintf('      csum(pr) = %18.16e\n', csum(pr));
fprintf(' csum(abs(pr)) = %18.16e\n', csum(abs(pr)));
fprintf('\n');

%%
% So at least our sums are coming out consistent.  We do not know what
% algorithm Matlab is using for compensated summation.  
%
% Of of curiosity sake, let's check the simple summation algorithm.  This
% code seems to indicate that Matlab is using just a simple (silly)
% summation algorithm.  

type ssum.m

fprintf('      csum(pr) = %18.16e\n', ssum(pr));
fprintf(' csum(abs(pr)) = %18.16e\n', ssum(abs(pr)));


%% Testing with big vectors
% To get a better sense of the state of summation in Matlab, let's test
% with some big vectors.
%
% We'll generate a vector that Matlab thinks has sum 1!

n=1e7;
x=rand(n,1);
x=x./sum(x);

fprintf('        sum(x) = %18.16e\n', sum(x));
fprintf('       ssum(x) = %18.16e\n', ssum(x));
fprintf('       csum(x) = %18.16e\n', csum(x));
fprintf('\n');

%% 
% None of the vectors have sum 1!
%
% Let's try the same thing with csum.

n=1e7;
x=rand(n,1);
x=x./csum(x);

fprintf('        sum(x) = %18.16e\n', sum(x));
fprintf('       ssum(x) = %18.16e\n', ssum(x));
fprintf('       csum(x) = %18.16e\n', csum(x));
fprintf('\n')

%%
% So we can accurately normalize a vector with csum, but not sum.  This
% example demonstrates one reason why we use the compensated summation
% algorithm inside the C++ code.

%% Performance
% Matlab implements a decent m file compiler in modern versions (2007+).
% Using this compiler, we can test the performance of the different
% versions.

n = 1e7;
ntrials = 15;

rand('state',0); dt=0; s=0;
for i=1:ntrials, x=rand(n,1); x=x./sum(x); tic, s=s+sum(x); dt=dt+toc; end
fprintf('      sum time : %6.3f sec;  s = %18.16e\n', dt, s);

rand('state',0); dt=0; s=0;
for i=1:ntrials, x=rand(n,1); x=x./sum(x); tic, s=s+ssum(x); dt=dt+toc; end
fprintf('     ssum time : %6.3f sec;  s = %18.16e\n', dt, s);

rand('state',0); dt=0; s=0;
for i=1:ntrials, x=rand(n,1); x=x./sum(x); tic, s=s+csum(x); dt=dt+toc; end
fprintf('     csum time : %6.3f sec;  s = %18.16e\n', dt, s);

