%% Show uncertainty in Morrison et al.s GeneRank matrix
% Morrison et al. define the GeneRank variant of PageRank to compute the
% important of genes from a set of gene expression measurement combined
% with a set of connectivity information about genes.  They report that
% alpha=0.75-0.85 gives the best results.  We'll work with their data an
% examine what happens for a large set of parameters in our mode.

%% Setup the experiment
% This experiment should be run from the rapr/experiments/generank
% directory
cwd = pwd;
dirtail = 'experiments/generank'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('rapr:dir','%s should be executed from rapr/%s\n',mfilename,dirtail);
end
addpath('../../matlab'); % ensure we have the RAPr codes available

%%
% Load the data.  The file generank.mat comes from 
% http://www.biomedcentral.com/1471-2105/6/233/additional/
% and then we converted the 4k by 4k matrix w_All to be a transition matrix
% for our codes.  
%
% However, we left the vector v = expr_data as is from the file, so we need
% to normalize it to be a probability vector in this code.  The
% normalization comes from the generank.m file available from the same
% webpage.  
load('../../data/generank.mat');
n=size(P,1); 

% normalize v to have sum one
v = abs(v); v = v/max(v);

%% Evaluate the algorithm for uniform measures
% For the GeneRank problem, the choice of A should not be driven by a
% random surfer model as in the PageRank case.  Instead, the choice is
% literally an unknown parameter alpha.  Consequently, a uniform
% distribution makes the most sense.  The authors of the GeneRank paper
% claim that 0.75 <= alpha <= 0.85 give the most interesting results.
% However, what should we pick inside this interval?
%
% In this experiment, we'll assume 

% set N for the gqrapr algorithm
N=50;

pts=0:.2:1; npts = length(pts); tic;
ktex=zeros(npts); ktstdx=ktex; % kendall tau ex,stdx
for pi=1:npts,for pj=pi+1:npts
    l=pts(pi); r=pts(pj); fprintf('starting [%3.1f,%3.1f]...',l,r); tic
    d=alphadist('unif',l,r);eA=d.mf(1);eA=eA(end);
    xeA=(speye(n)-eA*P')\v; xeA=xeA./norm(xeA,1); % compute x(E(A))
    [ex stdx] = gqrapr(P,50,d,'direct',v); % compute E[x(A)], Std[x(A)]
    ktex(pi,pj)=ktau(xeA,ex); ktstdx(pi,pj)=ktau(xeA,stdx); % compute taus
    fprintf (' ... done! %f secs\n', toc);
end, end

% save the output
save 'generank-unif.mat' pts ktex ktstdx

%%
% Display the output as a matlab table with colored cells.  This code
% generates the input to a table for the paper.  It colors the Matlab cells
% with the ktau strength.  Positive values generate red cells and negative
% values generate blue cells.  Unfortunately, we don't see any negative
% examples in these cases.  We generate two tables, one for ex, and one for
% stdx.  

load 'generank-unif.mat'; npts=length(pts);

ccv = @(x) [1 2-x 2-x].*[1 0.5 0.5]*(x>=0)+... %  red for pos
    [x+2 x+2 1].*[0.5 0.5 1]*(x<0);            % blue for neg
ccc = @(x) sprintf('\\cellcolor[rgb]{%0.2f,%0.2f,%0.2f}',ccv(x));

fprintf('ktau dist between x(E(A)) and E(x(A)) for A uniform \n');
disp(ktex); fprintf('... latex table code ... \n'); 

vals = ktex; fprintf(' & %1.1f', pts(2:end)); fprintf('\\\\ \\hline \n'); 
for pi=1:npts
    fprintf('%1.1f & ', pts(pi));
    for pj=1:npts
        if pj>pi, fprintf('%s %5.3f', ccc(vals(pi,pj)), vals(pi,pj)); end
        if pj==npts, fprintf('\\\\ \n'); elseif pj>1, fprintf(' & '); end
    end
end, fprintf('\n');

fprintf('ktau dist between x(E(A)) and Std(x(A)) for A uniform \n');
disp(ktstdx); fprintf('... latex table code ... \n'); 

vals = ktstdx; fprintf(' & %1.1f', pts(2:end)); fprintf('\\\\ \\hline \n'); 
for pi=1:npts
    fprintf('%1.1f & ', pts(pi));
    for pj=1:npts
        if pj>pi, fprintf('%s %5.3f', ccc(vals(pi,pj)), vals(pi,pj)); end
        if pj==npts, fprintf('\\\\ \n'); elseif pj>1, fprintf(' & '); end
    end
end, fprintf('\n');



%% Evaluate the algorithm for a set of beta measures
% All of our codes work for a beta distribution as well.  In this case,
% we'll look at a beta distribution over the interval [0,1].  Using an
% appropraite beta function allows us to heavily weight choices in the
% interval [0.75,0.85], but still consider values outside that interval.  A
% Beta(4,13) puts even approximately weight on [0.75-0.85] as on
% [0.65,0.75] and isn't a bad guestimate of this range.  In this
% experiment, we'll test a range of beta distributions for values of a,b
% less than 16.  

% set N for the gqrapr algorithm
N=50;

pts=1:3:16; npts = length(pts); tic;
ktex=zeros(npts); ktstdx=ktex; % kendall tau ex,stdx
for pi=1:npts, for pj=1:npts
    a=pts(pi); b=pts(pj); fprintf('starting (%i,%i)...',a,b); tic
    d=alphadist('beta',a,b);eA=d.mf(1);eA=eA(end);
    xeA=(speye(n)-eA*P')\v; xeA=xeA./norm(xeA,1); % compute x(E(A))
    [ex stdx] = gqrapr(P,50,d,'direct',v); % compute E[x(A)], Std[x(A)]
    ktex(pi,pj)=ktau(xeA,ex); ktstdx(pi,pj)=ktau(xeA,stdx); % compute taus
    fprintf (' ... done! %f secs\n', toc);
end, end

save 'generank-beta.mat' pts ktex ktstdx

%%
% Now write the tables, using the same color codes as before.  This time,
% we'll write the entire table instead of skipping the lower half.
load 'generank-beta.mat'; npts=length(pts);

fprintf('ktau dist between x(E(A)) and E(x(A)) for A beta \n');
disp(ktex); fprintf('... latex table code ... \n'); 

vals=ktex; fprintf(' & %i', pts); fprintf('\\\\ \\hline \n'); 
for pi=1:npts
    fprintf('%i & ', pts(pi));
    for pj=1:npts
        fprintf('%s %5.3f', ccc(vals(pi,pj)), vals(pi,pj));
        if pj==npts, fprintf('\\\\ \n'); else, fprintf(' & '); end
    end
end, fprintf('\n');

fprintf('ktau dist between x(E(A)) and Std(x(A)) for A beta \n');
disp(ktstdx); fprintf('... latex table code ... \n'); 

vals = ktstdx; fprintf(' & %i', pts); fprintf('\\\\ \\hline \n'); 
for pi=1:npts
    fprintf('%i & ', pts(pi));
    for pj=1:npts
        fprintf('%s %5.3f', ccc(vals(pi,pj)), vals(pi,pj));
        if pj==npts, fprintf('\\\\ \n'); else, fprintf(' & '); end
    end
end, fprintf('\n');