%% The eigensystem of the polynomial chaos expansions
% Theoeretically, we can show that the PCE eigensystem of dimension $N
% \times n$ can be solved in $N$ solutions of a linear system.  In this
% experiment, we demonstrate this theory.

%% Setup the experiment
addpath('../../matlab');
addpath('../../libbvg');
global datadir
if isempty(datadir), datadir = '../../data'; end
load([datadir filesep 'example-small.mat']);
n = size(P,1); v = ones(n,1)./n; d = v-P*v > 1./(2*n);
% save a copy of the matrix with the dangling node adjustment
Pd = P+sparse(d)*sparse(v)';

%% Generate the coefficients of the polynomial chaos system
% First, we solve the standard PCE problem.  Some of the output (which we
% get for free) gives the interpreted parameters of the problem.  We'll use
% this additional output to show our eigensystem properties.
bigN = 5;
[ex stdx coeffs T u h2] = pcerapr(Pd,bigN-1,alphadist('beta',1,2),'direct');
ccoeffs = (kron(speye(bigN),speye(n))-kron(T,Pd'))\kron(u,v);
norm(coeffs-reshape(ccoeffs,n,bigN),1)


%% Modify the PCE system
% Given the output from pcerapr, we have to solve 
% $$ [I_N\otimesI - T\otimesP]\rvec(X) = u \otimes v $$
% which we can modify to
% $$ [[I_N\otimesI - \Lambda \otimes P]\rvec(Y) = q \otimes v $$
% where $TX = X \Lambda$, $Z^T Y = X$ and $Zq = u$.

% first compute the eigenvalues and vectors of T
[Z L] = eig(T); 
Z(:,1) = Z(:,1)*sign(Z(1,1));
% the only key element is the vector $q$, let's compute it!
q = Z\u;

% solve the system and change the basis
y=(kron(speye(bigN),eye(n))-kron(L,Pd'))\kron(q,v);
Y=reshape(y,n,bigN);
X=Y*Z';

% check the solution accuracy
X-coeffs

%% Evaluate problems
d = zeros(bigN,1);
d(1) = 2; for i=2:bigN, d(i) = sqrt(d(i-1)^2*T(i-1,i)/T(i,i-1)); end

Tt = diag(d)*T*diag(1./d);
[Z L] = eig(Tt);
Z(:,1) = Z(:,1)*sign(Z(1,1));
Zd=diag(1./d)*Z;

q=Z\u;