%% Updated Jacobi iteration for the PCE Linear System
% The PCE linear system solves is diagonally dominant and solves nicely
% with the Jacobi method.  Because the system has some block structure, it
% would be tempting to try a Block Jacobi method.  Unfortunately, this
% involves a PageRank solve on each iteration so it is necessarily slower
% than a Gaussian Quadrature approach (one iteration is equivalent to a GQ
% run).  Instead, we may be able to save both time and memory if we perform
% an updated Jacobi iteration, where the most recent values from all
% previous blocks are used instead of the last set of values.  Some quick
% tests with this approach showed that it did not converge.  In this 

%% Setup the experiment
addpath('../../matlab');
addpath('../../libbvg');
global datadir
if isempty(datadir), datadir = '../../data'; end
load([datadir filesep 'example-small.mat']);
n = size(P,1); v = ones(n,1)./n; d = v-P*v > 0.5;
% save a copy of the matrix with the dangling node adjustment
Pd = P+sparse(d)*sparse(v)';

%% Get the PCE matrix and an accurate solution
[dex dstdx dcoeffs T u] = pcerapr(P,5,[],'direct');
bigN = size(T,1);
% construct the linear system
A = speye(bigN*n) - kron(T,Pd');
b = kron(u,v);
% construct the updated Jacobi iteration
M = speye(bigN*n) - kron(tril(T,-1),Pd');
N = kron(triu(T),Pd');

% test that we got it right
fprintf('||A-(M-N)||_1 = %15g\n', norm(A-M+N,1));

%% Check the eigenvalues of the Jacobi iteration matrix
% What we are calling Jacobi isn't the classical Jacobi method, but a more
% natural iterpretation of that method in light of the problem.  In
% particular, the two methods are equivalent if our matrix P has a non-zero
% diagonal.  We don't inforce that in any of our methods and treat it
% purely as an input parameter.

jevals = eig(full(kron(T,Pd))); jrho=max(abs(jevals));

figure(1); clf; hold on; plot(evals,'.'); title(sprintf('\\rho(T \\otimes P^T) = %g', jrho));
plot(cos(0:0.01:2*pi),sin(0:0.01:2*pi),'k'); axis square; box off; axis off;

%% Check the eigenvalues of the iteration matrix
% From the theory of stationary iterations, we have that the method is
% convergent if M^{-1} N has all its eigenvalues in the unit circle.  Let's
% test this possibilitiy.

evals = eig(full(M\N)); rho=max(abs(evals));

figure(2); clf; hold on; plot(evals,'.'); title(sprintf('\\rho(M^{-1} N) = %g', rho));
plot(cos(0:0.01:2*pi),sin(0:0.01:2*pi),'k'); axis square; box off; axis off;

%%
% Based on these values, we should expect the updated Jacobi iteration to
% converge about log(rho)/log(jrho) times as fast.  Let's check this value:
log(rho)/log(jrho)


%%
% That seems to test out... let's see what happens when we try and
% apply this iteration itself.
x = zeros(bigN*n,1); maxit = 10;
for i=1:maxit, y = M\(N*x+b); fprintf('res(%4i) = %10g\n', i, norm(x-y,1)); x=y; end

%%
% It works, so there was a bug in my code!



