/*
 * David Gleich
 * Copyright, Stanford University, 2008
 * 24 January 2008
 */

/**
 * @file mcrapr.cc
 * Implement a Monte Carlo sampling 
 */
 
/**
 * 24 January 2008
 * Initial version
 * 
 * 4 February 2008
 * Modified to make MPI optional
 * 
 * 6 February 2008
 * Changed to binary output
 */
 
#ifdef RAPR_MPI 
#include <mpi.h>
#else
typedef int MPI_Comm;
MPI_Comm MPI_COMM_WORLD=1;
#endif // RAPR_MPI

extern "C" {
#include "bvgraph.h"
}

#include <vector>
#include <iostream>
#include <sstream>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <boost/random.hpp>
#include <boost/nondet_random.hpp>
#include "random_device.cpp"
#include "beta_distribution.hpp"

#include "rapr_common.hpp"

enum alg_tag {
    pagerank_alg_power,
    pagerank_alg_inout,
};
enum dist_tag { 
    dist_beta, dist_beta2, dist_unif
};

void run_experiments(bvgraph *g, alg_tag alg, std::string outputfilename, double tol, int maxit, 
                     int nsamples, unsigned int seed, dist_tag dist, double distl, double distr, double dista, double distb);

int mpiprintf(MPI_Comm comm, const char format[], ...);
int mpisyncprintf(MPI_Comm comm, const char format[], ...);

void print_usage(std::ostream& cerr)
{
    int rank=0;

#ifdef RAPR_MPI    
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
#endif // RAPR_MPI
    
    // exit if you aren't the head node
    if (rank) { return; }
    
    using namespace std;
    
    cerr << "usage: mcrapr graphfile output alg tol maxit samples seed distname [distparams]" << endl;
    cerr << endl;
    cerr << "Solves the PageRank problem on a Boldi-Vigna graph file" << endl;
    cerr << "with uniform teleportation in the strongly preferential model." << endl;
    cerr << endl;
    cerr << "For any option except graphfile, specify _ to get the default value." << endl;
    cerr << endl;
    cerr << "graphfile: the base filename of a bvgraph file.  (e.g. cnr-2000)" << endl;
    cerr << "output: output filename for pagerank vector, \'none\' means no output" << endl;
    cerr << "alg: the PageRank computation algorithm [\'power\' or \'inout\']" << endl;
    cerr << "tol: the stopping tolerance for the 1-norm error [float from (0,Infinity])" << endl;
    cerr << "maxit: the maximum number of iterations [positive integer]" << endl;
    cerr << endl;
    cerr << "seed: the random number generator seed [<= for random, >=0 for deterministic]" << endl;
    cerr << "samples: the number of mc samples [positive integer]" << endl;
    cerr << "distname: the name of the random distribution [\'beta\', \'beta2\', or \'unif\']" << endl;
    cerr << endl;
    cerr << "defaults: " << endl;
    cerr << "  output = \'none\'; alg = \'power\'; tol = 1e-12; maxit = 10000" << endl;
    cerr << "  seed = <random>; samples = 1000; distname = \'beta\'" << endl;
    cerr << endl;
    cerr << "distparams are a set of parameters to control the probability distribution" << endl;
    cerr << "and depend on the choice of distribution" << endl;
    cerr << "" << endl;
    cerr << "For dist = \'beta\' or \'beta2\' : " << endl;
    cerr << "  a b l r, where for beta(a,b) on [l,r] " << endl;
    cerr << "For dist = \'unif\' : " << endl;
    cerr << "  l r, for a uniform on [l,r] " << endl;
    cerr << "defaults: a = 2, b = 16, l = 0, r = 1" << endl; 
    cerr << endl;
}

template <typename T>
void parse_arg(std::string arg, T& val)
{
    if (arg.compare("_") != 0) {
        std::stringstream ss(arg);
        ss >> val;
    }
}

unsigned int random_seed() {
    #ifdef __linux__
    boost::random_device dev;
    return dev();
    #else
    #error The non-deterministic random device is currently available on Linux only.
    #endif // __linux__ 
}

void write_header() 
{ 
    // print off the 
    for (int i=0; i<60; i++) { mpiprintf(MPI_COMM_WORLD, "%c", '='); }
    mpiprintf(MPI_COMM_WORLD, "\n");

#ifdef RAPR_MPI        
    mpiprintf(MPI_COMM_WORLD, "pmcrapr \n\n");
#else 
    mpiprintf(MPI_COMM_WORLD, "mcrapr \n\n");
#endif     
    mpiprintf(MPI_COMM_WORLD, "David Gleich and Paul Constantine\n");
    mpiprintf(MPI_COMM_WORLD, "Copyright, 2008\n");
    
    for (int i=0; i<60; i++) { mpiprintf(MPI_COMM_WORLD, "%c", '='); }
    mpiprintf(MPI_COMM_WORLD, "\n");

#ifdef RAPR_MPI
    {
        char name[MPI_MAX_PROCESSOR_NAME+1];
        int namelen, rank=0, size=1;
        MPI_Comm_size(MPI_COMM_WORLD, &size); 
        MPI_Comm_rank(MPI_COMM_WORLD, &rank); 
        MPI_Get_processor_name(name, &namelen);
        
        mpiprintf(MPI_COMM_WORLD, "nprocs = %i\n", size);
        
        mpisyncprintf(MPI_COMM_WORLD, "[%3i] %s running...\n", rank, name);
        
        MPI_Barrier(MPI_COMM_WORLD);
    }
#endif // RAPR_MPI
    
    
}

int main(int argc, char **argv)
{
    int rank=0; 
    
#ifdef RAPR_MPI    
    // initialize MPI
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif // RAPR_MPI
    
    write_header();
    
    using namespace std;

    if (argc < 2 || argc > 13) {
        print_usage(std::cerr);
        return (-1);
    }

    std::string graphfilename;
    
    enum alg_tag alg = pagerank_alg_power;
    enum dist_tag dist = dist_beta; 

    bool output = false;
    std::string outputfilename = "none";
    double tol = 1e-12;
    int maxit = 20000;
    int nsamples = 1000;
    int seedarg = -1;
    double distl=0, distr=1, dista=2, distb=16; 
    
    graphfilename = argv[1];

    if (argc > 2) {
        std::string outputarg = argv[2];
        if (outputarg.compare("none") == 0 || outputarg.compare("_") == 0) {
            // don't change the default
        }
        else {
            output = true;
            outputfilename = outputarg;
        }
    }

    if (argc > 3) {
        std::string algarg = argv[3];
        if (algarg.compare("_") == 0) {
            // don't change the default
        }
        else if (algarg.compare("power") == 0) {
            alg = pagerank_alg_power;
        } 
        else if (algarg.compare("inout") == 0) {
            alg = pagerank_alg_inout;
        }
    }

    if (argc > 4) { parse_arg(argv[4], tol); }
    if (argc > 5) { parse_arg(argv[5], maxit); }
    if (argc > 6) { parse_arg(argv[6], nsamples); }
    if (argc > 7) { parse_arg(argv[7], seedarg); }

    if (argc > 8) {
        std::string distarg = argv[8];
        if (distarg.compare("_") == 0) {
            // don't change the default
        }
        else if (distarg.compare("beta") == 0) {
            dist = dist_beta;
        }
        else if (distarg.compare("beta2") == 0) {
            dist = dist_beta2;
        } 
        else if (distarg.compare("unif") == 0) {
            dist = dist_unif;
        }  
    }
    
    if (dist == dist_unif) {
        if (argc > 9) { parse_arg(argv[9], distl); }
        if (argc > 10) { parse_arg(argv[10], distr); }
    } else {
        if (argc > 9) { parse_arg(argv[9], dista); }
        if (argc > 10) { parse_arg(argv[10], distb); }
        if (argc > 11) { parse_arg(argv[11], distl); }
        if (argc > 12) { parse_arg(argv[12], distr); }
    }

    // generate the random seed from a non-deterministic generate
    unsigned int seed;    
    if (seedarg < 0) {
        seed = random_seed();
    } else {
        seed = seedarg;
    }
        
    if (!rank) {
        cout << "Parameters: " << endl;
        cout << "   graphfile = " << graphfilename << endl;
        if (output) { cout << "  outputfile = " << outputfilename << endl; }
        else { cout << "      output = none" << endl; }
        cout << "         alg = "; 
        switch (alg) {
            case pagerank_alg_power: cout << "power"; break;
            case pagerank_alg_inout: cout << "inout"; break;
        }
        cout << endl;
        cout << "         tol = " << tol << endl;
        cout << "       maxit = " << maxit << endl;
        cout << "     samples = " << nsamples << endl;
        cout << "        seed = " << seed << endl;
        cout << "        dist = ";
        switch (dist) {
            case dist_beta: cout << "beta(" << dista << ", " << distb << ") on (" << distl << ", " << distr << ")"; break;
            case dist_beta2: cout << "beta2(" << dista << ", " << distb << ") on (" << distl << ", " << distr << ")"; break;
            case dist_unif: cout << "uniform on (" << distl << ", " << distr << ")"; break;
        }
        cout << endl;
    }

    bvgraph g;
    int rval;

    rval = bvgraph_load(&g, graphfilename.c_str(), (unsigned int)graphfilename.length(), 0);
    if (rval) {
        cerr << "error: " << bvgraph_error_string(rval) << endl;
        return (rval);
    }

    if (!rank) {
        cout << "Graph information: " << endl;
        cout << "  nodes = " << g.n << endl;
        cout << "  edges = " << g.m << endl;
    }
    
    run_experiments(&g, alg, outputfilename, tol, maxit, nsamples, seed, dist, distl, distr, dista, distb);
     
#ifdef RAPR_MPI         
    // finialize MPI
    MPI_Finalize();
#endif // RAPR_MPI    

    return (0);
}

void run_experiments(bvgraph *g, alg_tag alg, std::string outputfilename, double tol, int maxit, 
                     int nsamples, unsigned int seed, dist_tag dist, double distl, double distr, double dista, double distb)
{
    using namespace boost;
    
    int rank=0, size=1;
#ifdef RAPR_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Comm_size(MPI_COMM_WORLD,&size);
#endif // RAPR_MPI
    
    //
    // 1.  Generate the set of MC samples and distribute to all processors
    //     with an attempt to balance computation.
    //
    
    std::vector<double> myalphas;
    
    // compute and distribute the samples
    if (!rank) {
        std::vector<double> alphas(nsamples);
        mt19937 rng(seed);
        if (dist == dist_beta) {
            // the definition of the beta distribution is awfully ambiguious,
            //  in this case, we need to swap a and b.
            beta_distribution<> beta(distb+1, dista+1);
            variate_generator<mt19937&,beta_distribution<> > gen(rng,beta);
            for (int i=0; i<nsamples; i++) { alphas[i] = (distr-distl)*gen()+distl; }
        } else if (dist==dist_beta2) {
            beta_distribution<> beta(distb+1, dista+1);
            variate_generator<mt19937&,beta_distribution<> > gen(rng,beta);
            for (int i=0; i<nsamples; i++) { double a=gen(); alphas[i] = (distr-distl)*a*a +distl; }
        } else if (dist==dist_unif) {
            uniform_real<> unif(distl, distr);
            boost::variate_generator<mt19937&,uniform_real<> > gen(rng,unif);
            for (int i=0; i<nsamples; i++) { alphas[i] = gen(); }
        }
        
        // output the alphas if we should
        if (outputfilename.compare("none")!=0) {
            std::string alphafilename = outputfilename+".alphas";
            FILE *alphafile = fopen(alphafilename.c_str(), "wt");
            if (alphafile) {
                for (int i=0; i<nsamples; i++) {
                    fprintf(alphafile,"%18.16e\n", alphas[i]);
                }
                fflush(alphafile);
                fclose(alphafile);
            } else {
                mpiprintf(MPI_COMM_WORLD,"Error, cannot write to file %s\n", alphafilename.c_str());
            }   
            
        }
        
        // compute a vector of relative times for each sample
        std::vector<double> reltimes(nsamples);
        for (int i=0; i<nsamples; i++) {
            reltimes[i] = log(tol)/log(alphas[i]);
        }
        
        // compute the distribution of the values of alpha
        std::vector< std::vector<double> > alphasdist(size);
        std::vector< double > procweight(size);
        for (int i=0; i<nsamples; i++) 
        {
            int minloadp=0;
            double minload=procweight[0];
            for (int p=1; p<size; p++) {
                if (procweight[p] < minload) {
                    minloadp = p; 
                    minload=procweight[p];
                }
            }
            alphasdist[minloadp].push_back(alphas[i]);
            procweight[minloadp] += reltimes[i];
        }
        
        for (int p=0; p<size; p++) {
            mpiprintf(MPI_COMM_WORLD,
                "  processor %3i has %6i samples with %10.2f estimated multiplies\n", 
                p, alphasdist[p].size(), procweight[p]);
        } 
        
        // send out the sets of alphas
#ifdef RAPR_MPI        
        for (int p=1; p<size; p++) {
            int nalphasforproc = alphasdist[p].size();
            //printf("[%3i] sending for 1 int (%i) to %i...\n", rank, nalphasforproc, p);            
            MPI_Send(&nalphasforproc,1,MPI_INT,p,0,MPI_COMM_WORLD);
            //printf("[%3i] sending for %i doubles to %i...\n", rank, alphasdist[p].size(), p);
            MPI_Send(&alphasdist[p][0],alphasdist[p].size(),MPI_DOUBLE,p,0,MPI_COMM_WORLD);
        }
#endif // RAPR_MPI        
        // save the set for proc0
        myalphas = alphasdist[0];
    } else {
#ifdef RAPR_MPI        
        int mynalphas = 0;
        MPI_Status stat;
        // receive the count of alphas
        //printf("[%3i] waiting for 1 int from root...\n", rank);
        MPI_Recv(&mynalphas,1,MPI_INT,0,0,MPI_COMM_WORLD,&stat);
        
        myalphas.resize(mynalphas);
        //printf("[%3i] waiting for %i doubles from root...\n", rank, mynalphas);
        MPI_Recv(&myalphas[0],mynalphas,MPI_DOUBLE,0,0,MPI_COMM_WORLD,&stat);
#endif // RAPR_MPI     
    }
    
    // everyone has their values of alpha now!
    mpisyncprintf(MPI_COMM_WORLD, "[%3i] running %6i samples ... \n", rank, (int)myalphas.size());
    
    //
    // 2.  Compute the MC samples and aggregate them locally into a 
    //     mean and variance precursor
    //
    
    std::vector<double> prvec(g->n);
    std::vector<double> meanvec(g->n);
    std::vector<double> varvec(g->n);
    
    for (size_t ai=0; ai<myalphas.size(); ai++)
    {
        //printf("[%3i] %6i/%6i alpha=%6.4f \n", rank, (int)ai, (int)myalphas.size(), myalphas[ai]); fflush(stdout);
        double a = myalphas[ai];
        
        // initialize the PageRank vector with a value of 1/n;
        for (int i=0; i<g->n; i++) {
            prvec[i]=1.0/(double)g->n;
        }
    
        if (alg == pagerank_alg_power) {
            power_alg(g, a, tol, maxit, prvec);
        }
        else if (alg == pagerank_alg_inout) {
            inner_outer_alg(g, a, tol, maxit, prvec);
        }
        
        // update the mean and variance
        for (int i=0; i<g->n; i++) {
            //double delta = (1.0/(double)(ai+1))*meanvec[i] - prvec[i];
            
            meanvec[i] += prvec[i];
            //varvec[i] = varvec[i] + prvec[i]*prvec[i] + ((double)(ai+1)/(double)(ai+2))*delta*delta;
            if (ai>0) { 
                double delta = ((double)(ai+1))*prvec[i]-meanvec[i];
                varvec[i] += (1.0/((double)((ai+1)*(ai))))*delta*delta;
            }
        }
    } 
    
    //
    // 3.  Aggregate all the mean and variance data on the root
    //
    if (!rank) {
        int totalsamples = (int)myalphas.size();
        std::vector<double> varvec2(g->n);
        
#ifdef RAPR_MPI           
        for (int p=1; p<size; p++) {
            int samplesonp;
            MPI_Status stat;
            
            MPI_Recv(&samplesonp,1,MPI_INT,p,0,MPI_COMM_WORLD,&stat);
            MPI_Recv(&prvec[0],g->n,MPI_DOUBLE,p,0,MPI_COMM_WORLD,&stat);
            MPI_Recv(&varvec2[0],g->n,MPI_DOUBLE,p,0,MPI_COMM_WORLD,&stat);
            
            // aggregate the samples
            for (int i=0; i<g->n; i++) {
                double delta = ((double)samplesonp/(double)totalsamples)*meanvec[i] - prvec[i];
                meanvec[i] += prvec[i];
                varvec[i] += varvec2[i] + ((double)totalsamples/(double)(samplesonp*(totalsamples+samplesonp)))*delta*delta;
            } 
            totalsamples += samplesonp;
        }
#endif // RAPR_MPI
        
        for (int i=0; i<g->n; i++) {
            meanvec[i] = meanvec[i]/((double)(totalsamples));
            varvec[i] = sqrt(varvec[i]/((double)(totalsamples-1))); 
        }
    } else {
#ifdef RAPR_MPI           
        // send your data to the root
        int mynalphas = myalphas.size();
        MPI_Send(&mynalphas,1,MPI_INT,0,0,MPI_COMM_WORLD);
        MPI_Send(&meanvec[0],g->n,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
        MPI_Send(&varvec[0],g->n,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
#endif // RAPR_MPI        
    }
    
    //
    // 4.  Output the mean and variance
    //
    
    if (!rank) {
        if (outputfilename.compare("none")!=0) {
                
            std::string meanfilename = outputfilename + ".mean";
            std::string stdfilename = outputfilename + ".std";
            
            FILE *meanf = fopen(meanfilename.c_str(),"wt");
            if (meanf) {
                fwrite(&meanvec[0],sizeof(double),g->n,meanf);   
                // for (int i=0; i<g->n; i++) {
                //     fprintf(meanf,"%18.16e\n",meanvec[i]);
                // }
                fflush(meanf);
                fclose(meanf);
            } else {
                mpiprintf(MPI_COMM_WORLD,"Error, cannot write to file %s\n", meanfilename.c_str());
            }
            
            FILE *stdf = fopen(stdfilename.c_str(),"w");
            if (stdf) {
                fwrite(&varvec[0],sizeof(double),g->n,stdf);   
                // for (int i=0; i<g->n; i++) {
                //     fprintf(stdf,"%18.16e\n",varvec[i]);
                // }
                fflush(stdf);
                fclose(stdf);
            } else {
                mpiprintf(MPI_COMM_WORLD,"Error, cannot write to file %s\n", stdfilename.c_str());
            }
        }
    }
}

/* ===========================================================================
 * MPI Utilities
 * ======================================================================== */
 
int mpiprintf(MPI_Comm comm, const char format[], ...) 
{
#ifdef RAPR_MPI    
    int rank;
    MPI_Comm_rank(comm,&rank); 
    if (!rank) {
        va_list args;
        va_start(args, format);
        vprintf(format, args);
        va_end(args);
        fflush(stdout);
    }
#else
    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);
    fflush(stdout);
#endif // RAPR_MPI
    return (0);
}

int mpisyncprintf(MPI_Comm comm, const char format[], ...) 
{
#ifdef RAPR_MPI    
    int rank, size;
    int token=0;
    MPI_Comm_rank(comm,&rank);
    MPI_Comm_size(comm,&size);
    
    MPI_Barrier(comm);
    
    if (rank) {
        MPI_Status s;
        MPI_Recv(&token,1,MPI_INT,rank-1,0,comm,&s);
    }
#endif // RAPR_MPI
    // always execute this code
    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);
    fflush(stdout);
#ifdef RAPR_MPI    
    if (rank+1!=size) {
        MPI_Send(&token,1,MPI_INT,rank+1,0,comm);
    }

    MPI_Barrier(comm);
#endif // RAPR_MPI    
    
    return (0);
}
