/*
 * David Gleich
 * Copyright, Stanford University, 2008
 * 4 February 2008
 */

/**
 * @file pcerapr_beta.cc
 * Solve the random alpha PageRank model using a polynomial chaos expansion. 
 */
 
/**
 * 4 February 2008
 * Initial version
 * 
 * 6 February 2008
 * Implemented updated Jacobi iteration and added stdx computation.
 * Switched to binary output
 */
 
extern "C" {
#include "bvgraph.h"
}
 
#include <vector>
#include <iostream>
#include <sstream>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "rapr_common.hpp"

enum alg_tag {
    pcesys_alg_jacobi,
    pcesys_alg_ujacobi
}; 

// define the main computation routine!
void pce_rapr_beta(bvgraph *g, int N, alg_tag alg, 
                   std::string outputfilename, 
                   double tol, int maxit, 
                   double distl, double distr, double dista, double distb);

double log_beta(double a, double b);
double beta(double a, double b);
double log_gamma(double n);

/** Compute the beta function for real-valued parameters
 * 
 * This implementation of the beta function is based on a computation for the 
 * logarithm of the gamma function from Numerical Recipes in C++, 218--221.
 * 
 * @param a the value of a
 * @param b the value of b
 * @return the value of the beta function for a, b
 */
double log_beta(double a, double b) {
    return log_gamma(a)+log_gamma(b)-log_gamma(a+b);
}  
double beta(double a, double b) {
    return exp(log_beta(a,b));
}

/** Compute the logarithm of Euler's gamma function
 * 
 * This implementation is based on Numerical Recipes in C++, 218--221.
 * 
 * @param n the value of n for the Gamma function.
 * @return log(Gamma(n))
 */
double log_gamma(double n) {
    // see Numerical Recipes in C++, p. 219 and 
    // Lanczos, C. SIAM J. Num. Anal. (B), 1:86-96
    int j; double x, y, tmp, ser;
    static const double cof[6] = {76.18009172947146,-86.50532032941677,
        24.01409824083091,-1.231739572450155,
        0.1208650973866179e-2,-0.5395239384953e-5};
    y=x=n; tmp=x+5.5; tmp -= (x+0.5)*log(tmp);
    ser=1.000000000190015;
    for (j=0;j<=5;j++) ser += cof[j]/++y;
    return -tmp+log(2.5066282746310005*ser/x);
}

const char usage_message[] = \
"usage: pcerapr_beta graphfile size output alg tol maxit a b l r\n" 
"\n" 
"Solves the random alpha PageRank problem on a Boldi-Vigna graph file\n" 
"with uniform teleportation in the strongly preferential model.\n" 
"\n" 
"For any option except graphfile, specify _ to get the default value.\n" 
"\n" 
"graphfile: the base filename of a bvgraph file.  (e.g. cnr-2000)\n" 
"size: the size of the polynomial chaos expansion. [positive integer]\n" 
"\n" 
"output: output filename for pagerank vector, \'none\' means no output\n" 
"alg: the PCE system computation algorithm [\'jacobi\']\n" 
"tol: the stopping tolerance for the 1-norm error [float from (0,Infinity])\n" 
"maxit: the maximum number of iterations [positive integer]\n" 
"\n" 
"a: the a parameter of the beta [positive real]\n"
"b: the a parameter of the beta [positive real]\n"
"l: the left endpoint of the beta interval [real, 0 <= l < r <= 1]\n"
"r: the left endpoint of the beta interval [real, 0 <= l < r <= 1]\n"
"\n" \
"defaults: \n" \
"  size = 9; output = \'none\'; \n" \
"  alg = \'jacobi\'; tol = 1e-12; maxit = 10000; \n" \
"  beta_a = 2; beta_b = 16; beta_l = 0; beta_r = 1; \n"
"";

void print_usage(std::ostream& cerr)
{   
    std::cerr << usage_message << std::endl;
}

template <typename T>
void parse_arg(std::string arg, T& val)
{
    if (arg.compare("_") != 0) {
        std::stringstream ss(arg);
        ss >> val;
    }
}

void write_header() 
{ 
    for (int i=0; i<60; i++) { printf("%c", '='); }
    printf("\n");
    printf("pcerapr_beta \n\n");
    printf("David Gleich and Paul Constantine\n");
    printf("Copyright, 2008\n");
    for (int i=0; i<60; i++) { printf("%c", '='); }
    printf("\n");
}

int main(int argc, char **argv)
{
    write_header();
    
    using namespace std;
    
    if (argc < 2 || argc > 11) {
        print_usage(std::cerr);
        return (-1);
    }

    std::string graphfilename;
    
    enum alg_tag alg = pcesys_alg_jacobi;
    bool output = false;
    std::string outputfilename = "none";
    double tol = 1e-12;
    int maxit = 20000;
    int size = 9;
    double distl=0, distr=1, dista=2, distb=16; 
    
    graphfilename = argv[1];
    
    if (argc > 2) { parse_arg(argv[2], size); }

    if (argc > 3) {
        std::string outputarg = argv[3];
        if (outputarg.compare("none") == 0 || outputarg.compare("_") == 0) {
            // don't change the default
        }
        else {
            output = true;
            outputfilename = outputarg;
        }
    }
    if (argc > 4) {
        std::string algarg = argv[4];
        if (algarg.compare("_") == 0) {
            // don't change the default
        }
        else if (algarg.compare("jacobi") == 0) {
            alg = pcesys_alg_jacobi;
        } 
    }
    
    if (argc > 5) { parse_arg(argv[5], tol); }
    if (argc > 6) { parse_arg(argv[6], maxit); }
    if (argc > 7) { parse_arg(argv[7], dista); }
    if (argc > 8) { parse_arg(argv[8], distb); }
    if (argc > 9) { parse_arg(argv[9], distl); }
    if (argc > 10) { parse_arg(argv[10], distr); }
    
    cout << "Parameters: " << endl;
    cout << "   graphfile = " << graphfilename << endl;
    if (output) { cout << "  outputfile = " << outputfilename << endl; }
    else { cout << "      output = none" << endl; }
    cout << "    pce_size = " << size << endl;
    cout << "  pcesys_alg = "; 
    switch (alg) {
        case pcesys_alg_jacobi: cout << "jacobi"; break;
        case pcesys_alg_ujacobi: cout << "ujacobi"; break;
    }
    cout << endl;
    cout << "         tol = " << tol << endl;
    cout << "       maxit = " << maxit << endl;
    cout << "        dist = ";
    cout << "beta(" << dista << ", " << distb << ") on (" << distl << ", " << distr << ")";
    cout << endl;
    
    bvgraph g;
    int rval;

#ifdef RAPR_OFFLINE
    rval = bvgraph_load(&g, graphfilename.c_str(), (unsigned int)graphfilename.length(), -1);
#else 
    rval = bvgraph_load(&g, graphfilename.c_str(), (unsigned int)graphfilename.length(), 0);
#endif // RAPR_OFFLINE
        
    if (rval) {
        cerr << "error: " << bvgraph_error_string(rval) << endl;
        return (rval);
    }
    
    cout << "Graph information: " << endl;
    cout << "  nodes = " << g.n << endl;
    cout << "  edges = " << g.m << endl;

    pce_rapr_beta(&g, size, alg, outputfilename, tol, maxit, distl, distr, dista, distb);
    
    return (0);
}

/**Compute the coefficients of the tridiagonal PCE linear system.
 * T[0,...,k,...,N] = T(k+1,k)
 * T[N+1,...,k,...,2N+1] = T(k-N,k-N)
 * T[2N+2,...,k,...,3N+2] = T(k-(2N+1),
 */
void pce_beta_coefficients(int N, std::vector<double> &h2,  
    std::vector<double> &T, double *u, double a, double b, double l, double r)
{
    double s = (r-l), n=1.0, h2k=0.0;
    std::vector<double> gamma(N+1),nu(N+1),mu(N+1);
    nu[0] = (b-a)/(a+b+2.0); mu[0] = 0; gamma[0]=2.0/(a+b+2.0);
    for (int i=1; i<=N; i++, n++) { 
        gamma[i]=(2.0*(n+1.0)*(n+a+b+1.))/((2*n+a+b+1)*(2*n+a+b+2));
        nu[i] = (b*b - a*a)/((2.0*n+a+b)*(2.0*n+a+b+2));
        mu[i] = (2.0*(n+a)*(n+b))/((2.0*n+a+b)*(2.0*n+a+b+1.0));
    }
    u[0] = 1 - s*(b+1.0)/(a+b+2.0)-l; // u[0] = 1-E[A]
    u[1] = -mu[1]*(s/2.0)*(2.0+a+b+1.0)/((1.0+a)*(1.0+b));
    for (int i=0; i<=N; i++) {
        T[i+N+1]=(s/2.0 + l)+nu[i]*s/2;
        if (i>0) { T[i] = (s/2.0)*gamma[i-1]; }
        if (i<N) { T[2*N+2+i] = (s/2.0)*mu[i+1]; }
    }
    /*
     * h2=zeros(N+1,1); h2k=exp(betaln(a+1,b+1)+log(s)*(a+b+1)); h2(1)=h2k;
for k=1:N, 
    h2k=(2*(k-1)+a+b+1)*h2k*(k+a)*(k+b)/((2*k+a+b+1)*k*(k+a+b));h2(k+1)=h2k; 
end */
    h2k=exp(log_beta(a+1,b+1)+log(s)*(a+b+1)); h2[0]=h2k;
    n=1.0; for (int i=1; i<= N; i++, n++) {
        h2k=(2.*(n-1.)+a+b+1)*h2k*(n+a)*(n+1)/((2.*n+a+b+1)*n*(n+a+b)); h2[i]=h2k;
    }
    
}

/** Solve a PCE tri-diagonal linear system using the Jacobi method.
 * 
 * Technically, the algorithm is not the Jacobi method as the linear
 * system may retain a slight diagonal, however, it is effectively
 * the Jacobi method and has a rigorous convergence analysis.
 * 
 * The coefficients in T are stored according to the documentation for
 * the pce_beta_coefficients function.
 * 
 * 
 * @param g the bvgraph structure
 * @param N the size of the polynomial chaos expansion
 * @param T the coefficients of the tridiagonal linear system
 * @param u the coefficients of the right hand side
 * @param tol the stopping tolerance
 * @param maxit the maximum number of iterations
 * @param xs the g.n*(N+1) array of polynomial chaos coefficients.
 */
void pcesys_jacobi(bvgraph *g, int N, double *T, double *u, double tol, int maxit,
                   double *xs)
{
    size_t n=g->n, bigN = (N+1)*g->n;
    std::vector<double> ysmem((N+1)*n);
    double *ys = &ysmem[0];
    
    for (int i = 0; i<N+1; i++) {
        double *xi=xs+i*(N+1);
        if (i < 2) { set(xi,u[i]*1./(double)n,n); }
        else { set (xi,0.0, n); }
    }
    
    double delta=2; int iter=0;
    while (delta > tol && iter++ < maxit) {
        delta = 0.0;
        for (int i = 0; i < N+1; i++) {
            double *yi=ys+i*n;
            double *xi=xs+i*n;
            double *xp=xs+(i+1)*n;
            double *xm=xs+(i-1)*n;
            set(yi,0.0,n);
            
            if (i>0) { dangling_mult(g,xm,yi,T[i],n); }
            dangling_mult(g,xi,yi,T[i+(N+1)],n);
            if (i<N) { dangling_mult(g,xp,yi,T[i+(2*N+2)],n); }
            if (i<2) { shift(yi,u[i]*1./(double)n,n); }
            delta += diff_norm_1(xi,yi,n);
        }

        { double *temp; temp = xs; xs = ys; ys = temp; }

        printf("jacobi : iter = %6i ; delta = %10e\n", iter, delta);
    }
    if (xs == &ysmem[0]) { copy(xs,ys,bigN); }
    if (delta > tol) {
        printf("jacobi did not converge to %8e in %6i iterations\n",tol,maxit);
    }
}

/** Solve a PCE tri-diagonal linear system using the updated Jacobi method.
 * 
 * Technically, the algorithm is not the Jacobi method as the linear
 * system may retain a slight diagonal, however, it is effectively
 * the Jacobi method and has a rigorous convergence analysis.
 * 
 * The coefficients in T are stored according to the documentation for
 * the pce_beta_coefficients function.
 * 
 * 
 * @param g the bvgraph structure
 * @param N the size of the polynomial chaos expansion
 * @param T the coefficients of the tridiagonal linear system
 * @param u the coefficients of the right hand side
 * @param tol the stopping tolerance
 * @param maxit the maximum number of iterations
 * @param xs the g.n*(N+1) array of polynomial chaos coefficients.
 */
void pcesys_ujacobi(bvgraph *g, int N, double *T, double *u, double tol, int maxit,
                   double *xs)
{
    size_t n=g->n;
    std::vector<double> ysmem(n);
    double *y = &ysmem[0];
    
    for (int i = 0; i<N+1; i++) {
        double *xi=xs+i*(N+1);
        if (i < 2) { set(xi,u[i]*1./(double)n,n); }
        else { set (xi,0.0, n); }
    }
    
    double delta=2; int iter=0;
    while (delta > tol && iter++ < maxit) {
        delta = 0.0;
        for (int i = 0; i < N+1; i++) {
            double *xi=xs+i*n;
            double *xp=xs+(i+1)*n;
            double *xm=xs+(i-1)*n;
            set(y,0.0,n);
            
            if (i>0) { dangling_mult(g,xm,y,T[i],n); }
            dangling_mult(g,xi,y,T[i+(N+1)],n);
            if (i<N) { dangling_mult(g,xp,y,T[i+(2*N+2)],n); }
            if (i<2) { shift(y,u[i]*1./(double)n,n); }
            delta += diff_norm_1(xi,y,n);
            copy(y,xi,n);
        }
        printf("ujacobi : iter = %6i ; delta = %10e\n", iter, delta);
    }
    if (delta > tol) {
        printf("ujacobi did not converge to %8e in %6i iterations\n",tol,maxit);
    }
}

/** Compute the expectation and variance of the random alpha PageRank model
 * using a polynomial chaos expansion.
 * 
 * This function writes the output to outputfilename.mean, outputfilename.std,
 * outputfilename.coeffs, outputfilename.Tu.
 */
void pce_rapr_beta(bvgraph *g, int N, alg_tag alg, 
                   std::string outputfilename, 
                   double tol, int maxit, 
                   double distl, double distr, double dista, double distb)
{
    using namespace std;
    
    // assumptions
    // N >= 0
    size_t n=g->n;
    
    // 
    // 1.  Compute the coefficients of the pce linear system.
    //
    vector<double> T(3*(N+1)), h2(N+1);
    double u[2] = {0};
    pce_beta_coefficients(N,h2,T,&u[0],dista,distb,distl,distr);
    
    // output the Ts and us if we should
    if (outputfilename.compare("none")!=0) {
        std::string Tufilename = outputfilename+".Tu";
        FILE *Tufile = fopen(Tufilename.c_str(), "wt");
        if (Tufile) {
            fprintf(Tufile,"%18.16e   %18.16e   %18.16e  %18.16e\n", 0., 0., u[0], u[1]);
            for (int i=0; i<N+1; i++) {
                fprintf(Tufile,"%18.16e   %18.16e   %18.16e  %18.16e\n", 
                    T[i],T[i+N+1],T[i+2*N+2],h2[i]);
            }
            fflush(Tufile);
            fclose(Tufile);
        } else {
            printf("Error, cannot write to file %s\n", Tufilename.c_str());
        }   
    }    
    
    //
    // 2.  Solve the linear system via alg.
    //
    vector<double> coeffs(g->n*(N+1));
    if (alg==pcesys_alg_jacobi) {
        pcesys_jacobi(g,N,&T[0],u,tol,maxit,&coeffs[0]);
    }
    
    if (outputfilename.compare("none")!=0) {
        std::string filename = outputfilename+".coeffs";
        FILE *f = fopen(filename.c_str(), "wt");
        if (f) {
            for (int j=0; j<N+1; j++) {
                for (size_t i=0; i<n; i++) {
                    fprintf(f,"%18.16e\n", coeffs[i+n*j]);
                }
            }
            fflush(f);
            fclose(f);
        } else {
            printf("Error, cannot write to file %s\n", filename.c_str());
        }   
    }
    
    
    //
    // 3.  Compute ex, stdx
    //
    if (outputfilename.compare("none")!=0) {
        std::string filename = outputfilename+".mean";
        FILE *f = fopen(filename.c_str(), "wb");
        if (f) {
            fwrite(&coeffs[0],sizeof(double),n,f);
            // for (size_t i=0; i<n; i++) {
            //    fprintf(f,"%18.16e\n", coeffs[i]);
            //}
            fflush(f);
            fclose(f);
        } else {
            printf("Error, cannot write to file %s\n", filename.c_str());
        }   
    }
    
    if (outputfilename.compare("none")!=0) {
        // compute stdx
        std::vector<double> std(n,0);
        for (int i=1; i<=N; i++) {
            double *xi=&coeffs[i*n];
            for (size_t j=0; j<n; j++) {
                std[j]+=h2[i]/h2[0]*xi[j]*xi[j];
            }
        }
        for (size_t i=0; i<n; i++) { std[i] = sqrt(std[i]); }
        std::string filename = outputfilename+".std";
        FILE *f = fopen(filename.c_str(), "wb");
        if (f) {
            fwrite(&std[0],sizeof(double),n,f);
            // for (size_t i=0; i<n; i++) {
            //    fprintf(f,"%18.16e\n", std[i]);
            // }
            fflush(f);
            fclose(f);
        } else {
            printf("Error, cannot write to file %s\n", filename.c_str());
        }   
    }     
}

