function diststruct=alphadist(name,varargin)
% ALPHADIST Internal information about teleportation parameter distributions
%
% dinfo=alphadist('unif') models a Uniform distribution on [0,1].
% dinfo=alphadist('unif',l,r) models a Uniform distribution on [l,r].
%
% dinfo=alphadist('beta') models a Beta(1,1) distribution on [0,1].
% dinfo=alphadist('beta',a,b) models a Beta(a,b) distribution on [0,1].
% dinfo=alphadist('beta',a,b,l,r) models a Beta(a,b) distribution on [0,1].
%
% This function generates the input for the 'dist' parameter to the rapr
% computation functions.  It takes the description of the distribution and
% builds a structure that encodes necessary functions of the random
% variables.
%
% Example:
%   d=alphadist('beta'); hist(d.rnd(1000),100); 
%   d=alphadist('unif',0.3,0.6); hist(d.rnd(1000),100); 
%   d=alphadist('beta',2,16,0,0.95); hist(d.rnd(1000),100); 

% David Gleich and Paul Constantine
% Copyright, Stanford University, 2008

% 2008 February 1
% Initial version
%
% 11 February 2008
% Removed p power
% Fixed errfunc typo

diststruct = struct('mf', @errfunc, 'rnd', @errfunc, ...
                    'qrule', @errfunc, ...
                    'normorthprods', @errfunc, 'orthprods', @errfunc);

switch name
    case 'unif'
        [l r] = process_distoptions([0 1],varargin{:});
        diststruct.mf = @(N) unif_moments(N,l,r);
        diststruct.rnd = @(m) (r-l)*rand(m,1)+l;
        diststruct.qrule = @(N) unif_qrule(N,l,r);
        diststruct.orthprods = @(N) beta_orthprods(N,0,0,l,r);
        diststruct.normorthprods = @(N) beta_normorthprods(N,0,0,l,r);
        diststruct.name = sprintf('unif(%g,%g)',l,r);
    case 'beta'
        [a b l r] = process_distoptions([1 1 0 1],varargin{:});
        diststruct.mf = @(N) beta_moments(N,a,b,l,r);
        diststruct.rnd = @(m) (r-l)*betarnd(b+1,a+1,m,1)+l;
        diststruct.qrule = @(N) beta_qrule(N,a,b,l,r);
        diststruct.orthprods = @(N) beta_orthprods(N,a,b,l,r);
        diststruct.normorthprods = @(N) beta_normorthprods(N,a,b,l,r);
        diststruct.name = sprintf('beta(%g,%g,%g,%g)',a,b,l,r);
    otherwise
        error('alphadist:invalidArgument', ...
            'the distribution %s is not recognized', name);
end

end

function [varargout] = process_distoptions(v,varargin)
    % copy options from v and varargin unless an option is null
    varargout = num2cell(v);
    for i=1:min(length(v),length(varargin))
        if ~isempty(varargin{i}), varargout(i) = varargin(i); end
    end
end

function errfunc(varargin)
error('alphadist:notImplemented', ...
    'This function is not implemented for this distribution.');
end
