function [h2,T,u]=beta_normorthprods(N,a,b,l,r)
% BETA_NORMORTHPRODS Beta(a,b,l,r) coefficients for the PCE linear system
%
% [h2,T,u] = beta_normorthprods(N,a,b,l,r) returns the inner-products of
% orthogonal polynomials for the beta distribution.  Namely, given the
% Jacobi polynomials over (l,r), we compute
%
% h2(i) = int(P(x,i-1)^2*w(x),l,r) = <P(i-1),P(i-1)>
% T(i,j) = int(x*P(x,j-1)*P(x,i-1)*w(x),l,r)/h2(i) = <P(j-1),P(i-1)>/h2(i)
% u(i) = int((1-x)*P(x,j-1)*P(x,i-1)*w(x),l,r)/h2(i) = <1-x,P(i-1)>/h2(i)
%
% These are the inputs to the linear system formulation which yields the
% coefficients of the polynomial chaos expansion of a random variable.  
%
% See also BETA_ORTHPRODS
%
% Example:
%   [h2 T u] = beta_normorthprods(4,1,1,0,1)
%   [h2 T u] = beta_normorthprods(4,0,0,0,1) % get products for uniform

% David Gleich and Paul Constantine
% Copyright, Stanford University, 2008

% 2008 February 3
% Initial Version
%
% 2008 February 21
% Fixed error with a+b+1 = 0 in gamma and h2k

n=0:N; n1=1:N; s=(r-l); % set the range of n's and the scale parameter
% construct the recursion coefficients of the jacobi polynomials
gamma=[2/(a+b+2) (2.*(n1+1).*(n1+a+b+1))./((2*n1+a+b+1).*(2*n1+a+b+2))];
nu=[(b-a)/(a+b+2) (b^2 - a^2)./((2*n1+a+b).*(2*n1+a+b+2))];
mu=[0 (2*(n1+a).*(n1+b))./((2*n1+a+b).*(2*n1+a+b+1))];

h2=zeros(N+1,1); h2k=exp(betaln(a+1,b+1)+log(s)*(a+b+1)); h2(1)=h2k;
if N>0, h2k=h2k*(1+a)*(1+b)/(2+a+b+1); h2(2) = h2k; end % form 
for k=2:N, 
    h2k=(2*(k-1)+a+b+1)*h2k*(k+a)*(k+b)/((2*k+a+b+1)*k*(k+a+b));h2(k+1)=h2k; 
end

u=zeros(N+1,1);u(1) = 1-(s*(b+1)/(a+b+2)+l); % u(1) = 1-E[A]
if N>0, u(2) = -mu(2)*(s/2)*(2+a+b+1)/((1+a)*(1+b)); end % u(2)=-mu_2*ht_0^2/ht_1^2
T=zeros(N+1,N+1); 
for k=0:N, 
    T(k+1,k+1)=(s/2+l)+nu(k+1)*s/2;
    if k+1<=N, T(k+1,k+2)=(s/2)*mu(k+2); end
    if k>0, T(k+1,k)=(s/2)*gamma(k); end
end
