function [h2,J,u]=beta_orthprods(N,a,b,l,r)
% BETA_ORTHPRODS Beta(a,b,l,r) un-normalized orthogonal products
%
% [h2,T,u] = beta_orthprods(N,a,b,l,r) returns the inner-products of
% orthogonal polynomials for the beta distribution without normalization for
% the PCE linear system.  Namely, given the Jacobi polynomials over (l,r), 
% it computes
%
% h2(i) = int(P(x,i-1)^2*w(x),l,r) = <P(i-1),P(i-1)>
% T(i,j) = int(x*P(x,j-1)*P(x,i-1)*w(x),l,r)/h2(i) = <P(j-1),P(i-1)>
% u(i) = int((1-x)*P(x,j-1)*P(x,i-1)*w(x),l,r)/h2(i) = <1-x,P(i-1)>
%
% These are the inputs to the linear system formulation which yields the
% coefficients of the polynomial chaos expansion of a random variable.  
%
% See also BETA_NORMORTHPRODS
%
% Example:
%   [h2 T u] = beta_orthprods(4,1,1,0,1)
%   [h2 nT nu] = beta_normorthprods(4,1,1,0,1) 
%   T - diag(h2)*nT
%   diag(h2)*nu

% David Gleich and Paul Constantine
% Copyright, Stanford University, 2008

% 2008 February 3
% Initial Version

n=0:N; n1=1:N; s=(r-l); % set the range of n's and the scale parameter
% construct the recursion coefficients of the jacobi polynomials
gamma=(2.*(n+1).*(n+a+b+1))./((2*n+a+b+1).*(2*n+a+b+2));
nu=[(b-a)/(a+b+2) (b^2 - a^2)./((2*n1+a+b).*(2*n1+a+b+2))];
mu=[0 (2*(n1+a).*(n1+b))./((2*n1+a+b).*(2*n1+a+b+1))];

h2=zeros(N+1,1); h2k=exp(betaln(a+1,b+1)+log(s)*(a+b+1)); h2(1)=h2k;
for k=1:N, 
    h2k=(2*(k-1)+a+b+1)*h2k*(k+a)*(k+b)/((2*k+a+b+1)*k*(k+a+b));h2(k+1)=h2k; 
end
J=zeros(N+1,N+1); 
for k=0:N, 
    J(k+1,k+1)=h2(k+1)*(s/2+l)+nu(k+1)*h2(k+1)*s/2;
    if k+1<=N, J(k+1,k+2)=(s/2)*(h2(k+1)*mu(k+2)); end
    if k>0, J(k+1,k)=(s/2)*(h2(k+1)*gamma(k)); end
end
u=zeros(N+1,1);u(1) = (1-(s*(b+1)/(a+b+2)+l))*h2k(1); % u(1) = 1-E[A]
if N>0, u(2) = -mu(2)*h2k(1); end % u(2)=-mu_1*ht_0^2=-mu_1