function [ex stdx xw X] = gqrapr(P,N,dist,tol,v)
% GQRAPR Compute the mean and stdev of the RAPR model via Gauss quadrature
%
% [ex stdx] = gqrapr(P,N) computes a Gaussian quadrature approximation to
% the integrals defining the expectation and standard deviation of the
% random alpha PageRank model using N PageRank vectors.  The alpha's are
% distributed according to Beta(1,1) and each PageRank system is solved to
% a weighted tolerance of 1e-12 with uniform teleportation in the strongly
% preferential model.  
%
% The output (ex) satisifes
%   ||ex_N - ex||_1 <= N*tol
% where ex_N is the accurate N-point Gauss quadrature approxiation.
%
% [ex stdx xw X] = gqrapr(P,N,dist,tol,v) specifies optional
% input parameters to use an alternate distribution (returned from
% ALPHADIST), adjust the PageRank solution tolerance, or adjust the 
% teleportation distribution vector.  The output xw gives the points and
% weights of the Gaussian quadrature rule.  If tol is 'direct', then we 
% use a direct method to compute the solution to each PageRank system.  The
% output X contains the solution of the PageRank system at each of the
% Gauss points.  It is only stored if requested.
%
% % See also MCRAPR, PDRAPR, PCERAPR
%
% Example:
%   load('../data/example-small.mat');
%   [ex stdx] = gqrapr(P,10);
%   [ex stdx] = gqrapr(P,20,alphadist('beta',2,16),'direct');

% Paul Constantine and David Gleich
% Copyright, Stanford University, 2008

% 2008 February 2
% Initial version

% Notes
% the stdx variable is actually E[x^2] until we normalize it ot be stdx at
% the end.

n = size(P,1);
if ~exist('v','var') || isempty(v), v = 1./n; end
if ~exist('tol','var') || isempty(tol), tol = 1e-12; end
if ~exist('dist','var') || isempty(dist), dist = alphadist('beta'); end
if ~isnumeric(tol) && ~strcmp(tol,'direct')
    error('gqrapr:invalidArgument','unknown option tol=%s\n', tol);
end
if ~isnumeric(tol), P=P'; v=zeros(n,1) + v; end % adjust for direct solves

if nargout>3, X=zeros(n,N); end;
xw = dist.qrule(N);                 % generate the quadrature rule
ex = zeros(n,1); stdx = zeros(n,1);
for i=1:N
    % solve the PageRank system
    if ~isnumeric(tol), x = (speye(n)-xw(i,1)*P)\v; x=x./norm(x,1);
    else x = inoutpr(P,xw(i,1),v,min(tol./xw(i,2),1e-2), ... % adjust tol and maxit
      2*ceil(log(min(tol./xw(i,2),1e-2))/log(xw(i,1))));  % for mult by xw(i,2)
    end
    if nargout>3, X(:,i)=x; end;
    ex = ex+xw(i,2).*x;
    stdx = stdx+xw(i,2).*(x.^2);
end
stdx = sqrt(stdx - ex.^2); % convert to stdx