function [ex stdx alphas conv] = mcrapr(P,N,dist,tol,v,es)
% MCRAPR Use Monte Carlo to approximate the mean and variance RAPr model
%
% [ex stdx] = mcrapr(P,nsamples) computes a Monte Carlo approximation to
% the mean and standard deviation of the Random Alpha PageRank model using
% a beta(1,1) distribution for alpha and solving the PageRank systems using
% the Power method to 1e-12 accuracy with a uniform teleportation vector.
%
% The input matrix P should be row stochastic or sub row stochastic.
%
% [ex stdx alphas conv] = mcrapr(P,nsamples,dist,tol,v) specifies alternate
% parameters.  The dist parameter must come from the alphadist function.  
% Set tol to 'direct' for direct solutions of each PageRank system.
% Set any parameter to the empty matrix to get its default value.    The
% values of alpha are the samples from the alpha distribution.  The diff
% vector stores the 1-norm difference ...
%
% See also PDRAPR, GQRAPR, PCERAPR
%
% Example:
%   load('../data/example-small.mat');
%   tic, [ex stdx] = mcrapr(P,10000,[],'direct'); toc
%   tic, [ex stdx] = mcrapr(P,10000); toc      % sometimes exact is faster!

% Paul Constantine and David Gleich
% Copyright, Stanford University, 2008

% 2008 February 1
% Initial version
%
% 2008 February 21
% Fixed bug with not saving convergence history
%
% 2008 February 27
% Added time monitoring

n=size(P,1);

if ~exist('dist','var') || isempty(dist), dist=alphadist('beta'); end
if ~exist('tol','var') || isempty(tol), tol=1e-12; end
if ~exist('v','var') || isempty(v), v=1./n; end
if ~isnumeric(tol) && ~strcmp(tol,'direct')
    error('mcrapr:invalidArgument','unknown option tol=%s\n', tol);
end
ues=1;if ~exist('es','var') || isempty(es),ues=0;end % (u)es=(use) exact solns
if ~isnumeric(tol), P=P'; v=zeros(n,1) + v; end % adjust for direct solves

cdiff=nargout>3; % i==ci(i) => store iteration, i==d(i) => compute difference
if cdiff, ci=unique(round(logspace(0,log10(N-2),50)))+1; d=ci+1; t0=clock; end
if cdiff, di=1; conv=zeros(numel(d),4); conv(:,1) = d; m=numel(d); end

alphas = dist.rnd(N);
ex=zeros(n,1); stdx=zeros(n,1);
for i=1:N
    % solve the PageRank system
    if ~isnumeric(tol), x = (speye(n)-alphas(i)*P)\v; x=x./norm(x,1);
    else x = inoutpr(P,alphas(i),v,tol,2*ceil(log(tol)/log(alphas(i))));
    end
    % update the running solution sum and variance sum formulas
    ex = ex+x;
    if i>1, stdx = stdx + (1./(i*(i-1))).*(i*x-ex).^2; end
    if cdiff && di<=m && i>=d(di), 
        conv(di,2)=norm(ex./i-pex,1);
        conv(di,3)=norm(sqrt(stdx./(i-1))-pstdx,1);
        conv(di,4)=etime(clock,t0); di=di+1;
    end
    % update the solution ues=0 => approx with iter, ues=1, use es
    if cdiff && di<=m && i>=ci(di) && ~ues, pex=ex./i;pstdx=sqrt(stdx./(i-1)); end
    if cdiff && di<=m && i>=ci(di) && ues, pex=es(:,1); pstdx=es(:,2); end
end

ex = ex./N; stdx=sqrt(stdx./(N-1)); % compute the mean and std
