function [ex stdx pcecoeffs T u h2] = pcerapr(P,N,dist,alg,tol,v)
% PCERAPR Solve the RAPR model via a polynomial chaos expansion
%
% [ex stdx] = pcerapr(P,N) generates an N degree polynomial chaos expansion
% for the PageRank random variables.  It solves an (N+1)*size(P,1)
% dimensional linear system to a tolerance of 1e-12 using the [[[fill in
% the final algorithm choice]]] 
%
% [ex stdx pcecoeffs T u] = pcrapr(P,N,dist,alg,tol,v) provides additional
% input parameters to use an alternate distribution (returned from
% ALPHADIST), adjust the PCE solution algorithm alg='direct', 'jacobi', 
% 'bjacobi', or 'cr' for a direct method (Matlab '\'), a Jacobi or Block
% Jacobi iterative method, or a cyclic reduction based method; tol changes
% the solution tolerance for the iterative algorithms ('jacobi, 'bjacobi', 
% 'cr'); and v changes the teleportation distribution vector.
%
% The output pcecoeffs provide the coefficients of the PCE expansion for
% the PageRank variables so that 
%   ex = pcecoeffs(:,1), and 
%   stdx = sum(pcecoeffs(:,2:end)).
% The matrix T and vector u gives the coefficients of the pce linear system
%   (I-kron(T,P+dv')*pcecoeffs(:)' = kron(u,v)
%
% See also MCRAPR, GQRAPR, PDRAPR
%
% Example:
%   load('../data/example-small.mat');
%   [ex stdx] = pcerapr(P,4);
%   [ex stdx] = pcerapr(P,9,alphadist('beta',2,16),'direct');

n = size(P,1);
if ~exist('v','var') || isempty(v), v = 1./n; end
if ~exist('tol','var') || isempty(tol), tol = 1e-12; end
if ~exist('dist','var') || isempty(dist), dist = alphadist('beta'); end
if ~exist('alg','var') || isempty(alg), alg = 'diag'; end

try
    [h2,T,u] = dist.normorthprods(N);
catch
    [h2,J,h2u] = dist.orthprods(N);
    T = diag(sparse(1./h2))*J;
    u = h2u./h2;
end

switch alg
    case 'direct', pcecoeffs = solve_pcesys_direct(P,T,u,v,tol,1); 
    case 'jacobi', pcecoeffs = solve_pcesys_jacobi(P,T,u,v,tol,10000);
    case 'ujacobi', pcecoeffs = solve_pcesys_ujacobi(P,T,u,v,tol,10000);
    case 'diag', pcecoeffs = solve_pcesys_diag(P,T,u,v,tol,10000);
end
ex = pcecoeffs(:,1);
if N>0, stdx = sqrt((pcecoeffs(:,2:end).^2)*(h2(2:end)./h2(1))); 
else stdx = Inf*ones(n,1); warning('stdx not computed when N=0'); end    