function [ex ex2 pdcoeffs conv] = pdrapr(P,dist,tol,maxterms,v,es)
% PDRAPR Compute the mean and stdev of the RAPr model by path damping
%
% [ex stdx] = pdrapr(P) compute the expectation of the random alpha
% PageRank model when the teleportation parameters are distribution
% according to Beta(1,1). The algorithm is based on the Neumann expansion
% of the PageRank vector and its interpretation as a path damping formula
% and stops when the change in ex is less than 1e-6 or the computation uses
% 10000 terms.  The teleportation distribution vector is uniform.
%
% The output satisfies
%   ex = sum_i=0^N pdcoeffs(i+1)*P^i*v + (1-w)*P^(N+1)*v
% where w is 1 - sum(pdcoeffs(1:end-1)) (namely, the sum of the rest of the
% series.)  Unless specifically enabled, this function will not compute
% stdx and return NaNs instead as well as throw a warning.  The value N is
% maxterms, or a value such that w is less than tol.  
%
% pdcoeffs(i) = E[A^i-1 - A^i] for the random variable A distributed
% accoridng to dist.
%
% [ex stdx pdcoeffs] = pdrapr(P,dist,tol,maxterms,v) specifies optional
% input parameters to use an alternate distribution (returned from
% ALPHADIST), adjust the stopping tolerance, adjust the maximum number of
% terms possible, or adjust the teleportation distribution vector.  
%
% Truncating the Neumann series is analogous to a path damping expression
% for PageRank where the weight of a path of n is pdcoeffs(n+1).
%
% NOTE: The computation for stdx with this function is extremely slow!  It
% should not be used except for purely academic examples.  Consequently,
% the function will not compute it unless the tolerance is negative.
%
% See also MCRAPR, GQRAPR, PCERAPR
% 
% Example:
%   load('../data/example-small.mat');
%   [ex stdx] = pdrapr(P);
%   [ex stdx] = pdrapr(P,[],-1e-6); % actually compute stdx


% Paul Constantine and David Gleich
% Copyright, Stanford University, 2008

% 2008 February 1
% Initial version
%
% 2008 February 27
% Added time monitoring

% Notes
% the output variable for stdx is ex2 so we don't use any extra memory
% making yet another copy of the variable and more things work in-place

n = size(P,1);
if ~exist('v','var') || isempty(v), v = 1./n; end
if ~exist('tol','var') || isempty(tol), tol = 1e-6; end
if ~exist('maxterms','var') || isempty(maxterms), maxterms = 10000; end
if ~exist('dist','var') || isempty(dist), dist = alphadist('beta'); end
ues=1;if ~exist('es','var') || isempty(es),ues=0;end % (u)es=(use) exact solns

% check if we are computing stdx
compute_stdx=0; if tol<0, compute_stdx=1; tol=abs(tol); end
% setup the moments for the path damping coefficients
ms = dist.mf((maxterms+1)+compute_stdx*(maxterms+1));
if nargout>2, pdcoeffs = ms(1:end-1)-ms(2:end); end

cdiff=nargout>3; % d is the index to compute the difference at
if cdiff, d=unique(round(logspace(0,log10(maxterms),50))); t0=clock; end
if cdiff, di=1; conv=zeros(length(d),3); conv(:,1)=d; end % initialize factors

i=0; delta=2; ex=zeros(n,1); y = zeros(n,1) + v; s=0; % setup vectors
while i<maxterms && ms(i+2)>tol
    Ptiv=y; ex = ex + (ms(i+1)-ms(i+2))*Ptiv;
    y = P'*(Ptiv); y = y + (1-norm(y,1)).*v; i=i+1; % update P^i v
    if cdiff && i>=d(di),
        conv(di,3)=etime(clock,t0);
        if ~ues, conv(di,2)=ms(i+1)*norm(Ptiv-y,1);di=di+1; 
        else conv(di,2)=norm(es(:,1)-ex+ms(i+2)*y);di=di+1; end, end
end
ex = ex + ms(i+2)*y; % adjust with the last term

if nargout>2, pdcoeffs=pdcoeffs(1:i+1); end % adjust the output for the terms used

if compute_stdx
    warning('pdrapr:stdx','computing stdx');
    nterms=i; ex2=zeros(n,1); Ptiv = zeros(n,1); Ptiv = Ptiv + v; Ptjv=Ptiv;
    for i=0:nterms
        for j=0:nterms
            ex2 = ex2 + (ms(i+j+1)-2*ms(i+j+2)+ms(i+j+3))*(Ptiv.*Ptjv);
            y = P'*(Ptjv); Ptjv = y + (1-norm(y,1)).*v;
        end % update Ptiv and reset Ptjv
        y = P'*(Ptiv); Ptiv = y + (1-norm(y,1)).*v; Ptjv(:)=0; Ptjv=Ptjv+v;
    end
    ex2 = sqrt(ex2-ex.^2);
elseif nargout>1 
    ex2=Inf*ones(n,1); warning('pdrapr:stdx','not computing stdx');
end

