%% Speed and Accuracy of Direct Methods for PageRank
% In evaluating the quadrature formulas to solve for the expected value of
% the PageRank vector exactly, we need to solve about 3000 PageRank linear
% systems.  We want to have each solution as accurate as possible.  In this
% experiment, we evaluate the accuracy of our solutions as well as the time
% they take.

%% Experimental setup
% This experiment should be run from the rapr/experiments/correctness
% directory.  To reduce the computational load, we'll load the cs-stanford
% graph and just take the largest strong component.
cwd = pwd;
dirtail = 'experiments/correctness'; 
if strcmp(cwd(end-length(dirtail)+1:end),dirtail) == 0
    warning('%s should be executed from rapr/%s\n', mfilename, dirtail);
end
load ('../../data/wb-cs.stanford.mat');
% just use the largest strong component for this experiment;
P = Pcc; A = Acc;
%load ('../../data/example-small.mat');
n=size(P,1);

%%
% Add the path 
addpath('../../matlab');

%% Generate a set of alphas
% One of the goals is to see how long each method takes: Hessenberg update
% (theoretically faster) vs. Sparse LU factorization (exploits sparsity)
% for a range of values of alpha.  The Hessenberg method should not be
% sensitive to this parameter, but the Sparse LU factorization may be due
% to the pivot tolerance.
%
% To generate the alpha's, we'll use the Chebyshev points computed for the
% Fejer type 2 rule.

as=0.5*(fejer2_rule_compute(15)+1);m=max(as);as=[as;(1-m)*as+m];nas=length(as);

%% Hessenberg solution
% We are going to compute the solutions for many values of alpha.  
% The computationally efficient means of updating
% alpha is to use an upper hessenberg decomposition of the matrix P.  
% This is real and always exists.  Then the solution can be computed in 
% O(n^2) time for any value a, with good numerical stability.

%% 
% Factorize the matrix once. 
[Q,H] = hess(full(P'));

%% 
% Generate the appropriate right hand side.
v = ones(n,1)./n; w = Q'*v;

%%
% Solve all the systems
tic, for ai=1:nas, x=solve_pr_hess(n,Q,H,as(ai),w); end, dt=toc;
fprintf('%40s : %7f sec\n', 'solving Hessenberg systems', dt);

%%
% One thing we need to check is if the function call itself is causing the
% overhead.  This next call is the same thing, but with everything
% un-abstracted into the body.
tic; sopts=struct('UHESS',true); % tell Matlab our system is hessenberg
for ai=1:nas
    a = as(ai);
    Ha = -a*H;
    Ha = H+eye(n);
    wa = (1-a)*w;
    x = linsolve(Ha,wa,sopts);
    x = Q*x;
    s = norm(x,1);
    x = x./s;
end
dt=toc;
fprintf('%40s : %7f sec\n', '(inline) solving Hessenberg systems', dt);    

%% Direct solution
% Now, we'll try the direct solution using a sparse LU factorization of the
% matrix (I-a*P') directly.  This function is implemented directly in Matlab
% as the "backslash" operation for sparse matrices.  The key issue here is
% how best to form the matrix (I-a*P').  We will try a few varieties.  

%%
% The first variety is the most obvious: construct everything in one line.
tic, for ai=1:nas, x=(speye(n)-as(ai)*P')\((1-as(ai))*v); x=x./norm(x,1); end, dt=toc;
fprintf('%40s : %7f sec\n', 'simple SparseLU', dt);

%%
% The second variety keeps the matrix transposed first
Pt = P';
tic, for ai=1:nas, x=(speye(n)-as(ai)*Pt)\((1-as(ai))*v);  x=x./norm(x,1); end, dt=toc;
fprintf('%40s : %7f sec\n', 'pre-stored transpose SparseLU', dt);

%%
% The third variety keeps the matrix explicitly stores the rhs and matrix
Pt = P';
tic, for ai=1:nas, Pa=(speye(n)-as(ai)*Pt); va=v; va=(1-as(ai))*va; x=Pa\va; x=x./norm(x,1); end, dt=toc;
fprintf('%40s : %7f sec\n', 'explicit matrix rhs SparseLU', dt);

%%
% The fourth variety keeps the matrix explicitly stores the rhs
Pt = P';
tic, for ai=1:nas, va=v; va=(1-as(ai))*va; x=(speye(n)-as(ai)*Pt)\va; x=x./norm(x,1); end, dt=toc;
fprintf('%40s : %7f sec\n', 'explicit rhs SparseLU', dt);

%% Speed conclusion
% The fastest way of going is definitely the SparseLU approach.  The
% simplest optimization is just storing the matrix transpose explicitly.
% This yields a tiny optimization.

%% Accuracy
% One of the problems with the SparseLU approach is that the accuracy
% bounds may not be as good as for the Hessenberg case.  We will test this
% idea by first, looking at the difference in the solution between the
% Hessenberg and SparseLU solutions and checking the residuals.

for ai=1:nas
    a=as(ai);
    xlu = (speye(n)-a*Pt)\((1-a)*v); xlu=xlu./norm(x,1);
    xh = solve_pr_hess(n,Q,H,a,w);
    rlu = Pt*xlu; rlu = xlu - a*(rlu + (sum(xlu)-sum(rlu)).*v) - (1-a).*v;
    rh = Pt*xh; rh = xh - a*(rh + (sum(xh)-sum(rh)).*v) - (1-a).*v;    
    fprintf('a = %7.5f  ||xlu-xh|| = %10e  r(xlu) = %10e  r(xh) = %10e\n', ...
        a, norm(xlu-xh,1), norm(rlu,1), norm(rh,1));
end
    
%% 
% Unfortunately, these two solutions start to disagree as alpha -> 1.  This
% fact is not surprising as the problem becomes very illconditioned at that
% point.  Also, the residual doesn't tell the full tail.  To fix the
% situation, we are going to explicitly compute some right hand sides from
% known values of the solution.

nrhs=5;
for ai=1:nas
    a = as(ai);
    for rhsi=1:nrhs
        y = rand(n,1); y=y./norm(y,1);
        z = Pt*y; z = y - a*z;
        xlu = (speye(n)-a*Pt)\z;
        xh = solve_pr_hess(n,Q,H,a,Q'*(z./(1-a)));
        fprintf('a = %7.5f  t=%i  e(xlu) = %10e  e(xh) = %10e\n',...
            a, rhsi, norm(xlu-y,1), norm(xh-y,1));
    end
end
        
%% SparseLU with iterative refinement
% The previous experiment showed that SparseLU was generally more accurate
% than the Hessenberg solution.  Because the answers are so much faster, we
% could afford a few steps of iterative refinement.

nrhs=5;
for ai=1:nas
    a = as(ai);
    for rhsi=1:nrhs
        y = rand(n,1); y=y./norm(y,1);
        z = Pt*y; z = y - a*z;
        xlu = (speye(n)-a*Pt)\z;
        rlu = z-xlu+a*(Pt*xlu);
        z1 = (speye(n)-a*Pt)\rlu;
        xlu1 = xlu+z1;
        rlu2 = z-xlu1+a*(Pt*xlu1);
        z2 = (speye(n)-a*Pt)\rlu2;
        xlu2 = xlu1+z2;
        fprintf('a = %7.5f  t=%i  e(xlu) = %10e  e(xlu1) = %10e  e(xlu2) = %10e\n',...
            a, rhsi, norm(xlu-y,1), norm(xlu1-y,1), norm(xlu2-y,1));
    end
end

%%
% Iterative refinement didn't seem to help much.  My conclusion from this
% experiment is that just using the standard backslash solver is best
% (R2007a/b) as long as the total quadrature weight on the nodes above 0.99
% is less than 1e-4